"""Progress reporter for sending evaluation updates to StudioWeb."""

import functools
import json
import logging
import os
import uuid
from typing import Any
from urllib.parse import urlparse

from opentelemetry import trace
from pydantic import BaseModel
from rich.console import Console

from uipath._cli._evals._models._evaluation_set import (
    EvaluationItem,
    EvaluationStatus,
)
from uipath._cli._evals._models._sw_reporting import (
    StudioWebAgentSnapshot,
    StudioWebProgressItem,
)
from uipath._cli._utils._console import ConsoleLogger
from uipath._events._event_bus import EventBus
from uipath._events._events import (
    EvalRunCreatedEvent,
    EvalRunUpdatedEvent,
    EvalSetRunCreatedEvent,
    EvalSetRunUpdatedEvent,
    EvaluationEvents,
)
from uipath._utils import Endpoint, RequestSpec
from uipath._utils.constants import (
    ENV_EVAL_BACKEND_URL,
    ENV_TENANT_ID,
    HEADER_INTERNAL_TENANT_ID,
)
from uipath.eval.evaluators import (
    BaseEvaluator,
    BaseLegacyEvaluator,
)
from uipath.eval.evaluators.base_evaluator import GenericBaseEvaluator
from uipath.eval.models import EvalItemResult, ScoreType
from uipath.platform import UiPath
from uipath.platform.common import UiPathConfig

logger = logging.getLogger(__name__)


def gracefully_handle_errors(func):
    """Decorator to catch and log errors without stopping execution."""

    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        try:
            return await func(self, *args, **kwargs)
        except Exception as e:
            if hasattr(self, "_console"):
                error_type = type(e).__name__
                # Log the full error message for debugging
                logger.debug(f"Full error details: {e}")
                logger.warning(
                    f"Cannot report progress to SW. "
                    f"Function: {func.__name__}, "
                    f"Error type: {error_type}, "
                    f"Details: {e}"
                )
            return None

    return wrapper


class StudioWebProgressReporter:
    """Handles reporting evaluation progress to StudioWeb."""

    def __init__(self):
        logging.getLogger("uipath._cli.middlewares").setLevel(logging.CRITICAL)
        console_logger = ConsoleLogger.get_instance()

        # Use UIPATH_EVAL_BACKEND_URL for eval-specific routing if set
        eval_backend_url = os.getenv(ENV_EVAL_BACKEND_URL)
        uipath = UiPath(base_url=eval_backend_url) if eval_backend_url else UiPath()

        self._client = uipath.api_client
        self._console = console_logger
        self._rich_console = Console()
        self._project_id = os.getenv("UIPATH_PROJECT_ID", None)
        if not self._project_id:
            logger.warning(
                "Cannot report data to StudioWeb. Please set UIPATH_PROJECT_ID."
            )

        self.eval_set_ids: dict[str, str] = {}  # Track eval_set_id per execution
        self.eval_set_run_ids: dict[str, str] = {}
        self.evaluators: dict[str, Any] = {}
        self.evaluator_scores: dict[str, list[float]] = {}
        self.eval_run_ids: dict[str, str] = {}
        self.is_coded_eval: dict[str, bool] = {}  # Track coded vs legacy per execution
        self.is_resume_mode: dict[str, bool] = {}  # Track resume mode per execution
        self.eval_spans: dict[
            str, list[Any]
        ] = {}  # Store spans per execution for usage metrics
        self.eval_set_execution_id: str | None = (
            None  # Track current eval set execution ID
        )
        self.user_provided_eval_set_run_ids: set[str] = (
            set()
        )  # Track user-provided eval_set_run_ids

    @gracefully_handle_errors
    async def get_eval_run_for_evaluation(
        self,
        eval_set_id: str,
        eval_set_run_id: str,
        evaluation_id: str,
        is_coded: bool = False,
    ) -> str | None:
        """Get the eval_run_id for a specific evaluation from the backend.

        This is used during resume to fetch the eval_run_id for the specific
        evaluation being executed, using the backend database as the source of truth.

        Args:
            eval_set_id: The eval set ID
            eval_set_run_id: The eval set run ID
            evaluation_id: The specific evaluation ID being executed
            is_coded: Whether this is a coded evaluation (vs legacy)

        Returns:
            The eval_run_id if found, None otherwise
        """
        logger.info(
            f"Fetching eval runs from backend: eval_set_id={eval_set_id}, "
            f"eval_set_run_id={eval_set_run_id}, evaluation_id={evaluation_id}, coded={is_coded}"
        )

        spec = self._get_eval_runs_spec(
            eval_set_id, eval_set_run_id, evaluation_id, is_coded
        )

        logger.debug(f"GET request endpoint: {spec.endpoint}")

        response = await self._client.request_async(
            method=spec.method,
            url=spec.endpoint,
            params=spec.params,
            headers=spec.headers,
            scoped="org" if self._is_localhost() else "tenant",
        )

        logger.info(
            f"GET eval runs response: status_code={response.status_code}, "
            f"content_length={len(response.content)} bytes"
        )

        # Parse response to find the eval run matching this evaluation_id
        response_data = json.loads(response.content)
        logger.debug(
            f"GET eval run response data for evaluation_id={evaluation_id}: {json.dumps(response_data, indent=2)}"
        )

        # Extract eval runs from response
        # Response format may vary between coded and legacy APIs
        eval_runs = (
            response_data
            if isinstance(response_data, list)
            else response_data.get("value", [])
        )

        logger.info(
            f"Backend returned {len(eval_runs)} eval run(s) for eval_set_run_id={eval_set_run_id}"
        )

        # Find the eval run that matches our evaluation_id
        for idx, eval_run in enumerate(eval_runs):
            eval_snapshot = eval_run.get("evalSnapshot", {})
            snapshot_eval_id = str(eval_snapshot.get("id", ""))
            eval_run_id_in_response = eval_run.get("id")

            logger.debug(
                f"Checking eval run [{idx}]: eval_run_id={eval_run_id_in_response}, "
                f"snapshot_eval_id={snapshot_eval_id}, target_evaluation_id={evaluation_id}"
            )

            if snapshot_eval_id == evaluation_id:
                eval_run_id = eval_run.get("id")
                if eval_run_id:
                    logger.info(
                        f"✓ MATCH FOUND: eval_run_id={eval_run_id} matches evaluation_id={evaluation_id} (resume scenario)"
                    )
                    return eval_run_id
                else:
                    logger.warning(
                        f"Found matching eval snapshot with evaluation_id={evaluation_id} but eval_run_id is missing in response"
                    )

        logger.warning(
            f"✗ NO MATCH: No eval run found in backend for evaluation_id={evaluation_id}. "
            f"Searched {len(eval_runs)} eval run(s) in eval_set_run_id={eval_set_run_id}. "
            f"Available evaluation IDs: {[str(er.get('evalSnapshot', {}).get('id', '')) for er in eval_runs]}"
        )
        return None

    async def fetch_and_cache_eval_runs(
        self,
        eval_set_id: str,
        eval_set_run_id: str,
        is_coded: bool = False,
    ) -> None:
        """Fetch all eval runs from backend and populate cache.

        This is used during resume to pre-populate the eval_run_id cache with
        all existing eval runs from the backend database.

        Args:
            eval_set_id: The eval set ID
            eval_set_run_id: The eval set run ID
            is_coded: Whether this is a coded evaluation (vs legacy)
        """
        logger.info(
            f"🔄 RESUME FLOW: Fetching all eval runs from backend to populate cache: "
            f"eval_set_id={eval_set_id}, eval_set_run_id={eval_set_run_id}, coded={is_coded}"
        )

        # Use empty evaluation_id to fetch all eval runs (not filtering by specific evaluation)
        spec = self._get_eval_runs_spec(
            eval_set_id, eval_set_run_id, evaluation_id="", is_coded=is_coded
        )

        logger.debug(f"GET all eval runs endpoint: {spec.endpoint}")

        response = await self._client.request_async(
            method=spec.method,
            url=spec.endpoint,
            params=spec.params,
            headers=spec.headers,
            scoped="org" if self._is_localhost() else "tenant",
        )

        logger.info(
            f"GET all eval runs response: status_code={response.status_code}, "
            f"content_length={len(response.content)} bytes"
        )

        # Parse response to extract all eval runs
        response_data = json.loads(response.content)

        # Extract eval runs from response
        # Response format may vary between coded and legacy APIs
        eval_runs = (
            response_data
            if isinstance(response_data, list)
            else response_data.get("value", [])
        )

        logger.info(
            f"✓ Backend returned {len(eval_runs)} eval run(s) for eval_set_run_id={eval_set_run_id}"
        )

        # Populate cache with all eval runs
        cached_count = 0
        for eval_run in eval_runs:
            eval_snapshot = eval_run.get("evalSnapshot", {})
            evaluation_id = str(eval_snapshot.get("id", ""))
            eval_run_id = eval_run.get("id")

            if evaluation_id and eval_run_id:
                # Cache using evaluation_id as the key
                # Since we don't have execution_id yet, we'll need to map by evaluation_id
                # Store in a temporary mapping that will be used when CREATE_EVAL_RUN would have fired
                self.eval_run_ids[evaluation_id] = eval_run_id
                cached_count += 1
                logger.debug(
                    f"✓ Cached eval_run_id={eval_run_id} for evaluation_id={evaluation_id}"
                )

        logger.info(
            f"✓ RESUME FLOW: Successfully cached {cached_count}/{len(eval_runs)} eval runs"
        )

    def _format_error_message(self, error: Exception, context: str) -> None:
        """Helper method to format and display error messages consistently."""
        self._rich_console.print(f"    • \u26a0  [dim]{context}: {error}[/dim]")

    def _is_localhost(self) -> bool:
        """Check if the eval backend URL is localhost.

        Returns:
            True if using localhost, False otherwise.
        """
        eval_backend_url = os.getenv(ENV_EVAL_BACKEND_URL, "")
        if eval_backend_url:
            try:
                parsed = urlparse(eval_backend_url)
                hostname = parsed.hostname or parsed.netloc.split(":")[0]
                return hostname.lower() in ("localhost", "127.0.0.1")
            except Exception:
                pass
        return False

    def _get_endpoint_prefix(self) -> str:
        """Determine the endpoint prefix based on environment.

        Checks UIPATH_EVAL_BACKEND_URL environment variable:
        - If set to localhost/127.0.0.1: returns "api/" (direct API access)
        - Otherwise: returns "agentsruntime_/api/" (service routing for alpha/prod)

        Returns:
            "api/" for localhost environments, "agentsruntime_/api/" for alpha/production.
        """
        if self._is_localhost():
            return "api/"
        return "agentsruntime_/api/"

    def _is_coded_evaluator(
        self, evaluators: list[GenericBaseEvaluator[Any, Any, Any]]
    ) -> bool:
        """Check if evaluators are coded (BaseEvaluator) vs legacy (LegacyBaseEvaluator).

        Args:
            evaluators: List of evaluators to check

        Returns:
            True if using coded evaluators, False for legacy evaluators
        """
        if not evaluators:
            return False
        # Check the first evaluator type
        return not isinstance(evaluators[0], BaseLegacyEvaluator)

    def _extract_usage_from_spans(
        self, spans: list[Any]
    ) -> dict[str, int | float | None]:
        """Extract token usage and cost from OpenTelemetry spans.

        Args:
            spans: List of ReadableSpan objects from agent execution

        Returns:
            Dictionary with tokens, completionTokens, promptTokens, and cost
        """
        total_tokens = 0
        completion_tokens = 0
        prompt_tokens = 0
        total_cost = 0.0

        for span in spans:
            try:
                # Handle both dictionary attributes and string Attributes field
                attrs = None
                if hasattr(span, "attributes") and span.attributes:
                    if isinstance(span.attributes, dict):
                        attrs = span.attributes
                    elif isinstance(span.attributes, str):
                        # Parse JSON string attributes
                        attrs = json.loads(span.attributes)

                # Also check for Attributes field (capitalized) from backend spans
                if not attrs and hasattr(span, "Attributes") and span.Attributes:
                    if isinstance(span.Attributes, str):
                        attrs = json.loads(span.Attributes)
                    elif isinstance(span.Attributes, dict):
                        attrs = span.Attributes

                if attrs:
                    # Try to get usage from nested usage object (backend format)
                    if "usage" in attrs and isinstance(attrs["usage"], dict):
                        usage = attrs["usage"]
                        prompt_tokens += usage.get("promptTokens", 0)
                        completion_tokens += usage.get("completionTokens", 0)
                        total_tokens += usage.get("totalTokens", 0)
                        # Cost might be in usage or at root level
                        total_cost += usage.get("cost", 0.0)

                    # Also try OpenTelemetry semantic conventions (SDK format)
                    prompt_tokens += attrs.get("gen_ai.usage.prompt_tokens", 0)
                    completion_tokens += attrs.get("gen_ai.usage.completion_tokens", 0)
                    total_tokens += attrs.get("gen_ai.usage.total_tokens", 0)
                    total_cost += attrs.get("gen_ai.usage.cost", 0.0)
                    total_cost += attrs.get("llm.usage.cost", 0.0)

            except (json.JSONDecodeError, AttributeError, TypeError) as e:
                logger.debug(f"Failed to parse span attributes: {e}")
                continue

        return {
            "tokens": total_tokens if total_tokens > 0 else None,
            "completionTokens": completion_tokens if completion_tokens > 0 else None,
            "promptTokens": prompt_tokens if prompt_tokens > 0 else None,
            "cost": total_cost if total_cost > 0 else None,
        }

    @gracefully_handle_errors
    async def create_eval_set_run_sw(
        self,
        eval_set_id: str,
        agent_snapshot: StudioWebAgentSnapshot,
        no_of_evals: int,
        evaluators: list[BaseLegacyEvaluator[Any]],
        is_coded: bool = False,
    ) -> str:
        """Create a new evaluation set run in StudioWeb."""
        spec = self._create_eval_set_run_spec(
            eval_set_id, agent_snapshot, no_of_evals, is_coded
        )
        response = await self._client.request_async(
            method=spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            headers=spec.headers,
            scoped="org" if self._is_localhost() else "tenant",
        )
        eval_set_run_id = json.loads(response.content)["id"]
        return eval_set_run_id

    @gracefully_handle_errors
    async def create_eval_run(
        self, eval_item: EvaluationItem, eval_set_run_id: str, is_coded: bool = False
    ) -> str | None:
        """Create a new evaluation run in StudioWeb.

        Args:
            eval_item: Dictionary containing evaluation data
            eval_set_run_id: The ID of the evaluation set run
            is_coded: Whether this is a coded evaluation (vs legacy)

        Returns:
            The ID of the created evaluation run
        """
        spec = self._create_eval_run_spec(eval_item, eval_set_run_id, is_coded)
        response = await self._client.request_async(
            method=spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            headers=spec.headers,
            scoped="org" if self._is_localhost() else "tenant",
        )

        # Parse response and extract eval_run_id
        response_data = json.loads(response.content)
        logger.debug(f"CREATE_EVAL_RUN response: {response_data}")

        eval_run_id = response_data.get("id")
        if not eval_run_id:
            logger.error(f"No 'id' field in CREATE_EVAL_RUN response: {response_data}")
            return None

        return eval_run_id

    @gracefully_handle_errors
    async def update_eval_run(
        self,
        sw_progress_item: StudioWebProgressItem,
        evaluators: dict[str, BaseEvaluator[Any, Any, Any]],
        is_coded: bool = False,
        spans: list[Any] | None = None,
    ):
        """Update an evaluation run with results."""
        coded_evaluators: dict[str, BaseEvaluator[Any, Any, Any]] = {}
        legacy_evaluators: dict[str, BaseLegacyEvaluator[Any]] = {}
        evaluator_runs: list[dict[str, Any]] = []
        evaluator_scores: list[dict[str, Any]] = []

        for k, v in evaluators.items():
            if isinstance(v, BaseLegacyEvaluator):
                legacy_evaluators[k] = v
            elif isinstance(v, BaseEvaluator):
                coded_evaluators[k] = v

        # Use coded evaluator format
        runs, scores = self._collect_coded_results(
            sw_progress_item.eval_results, coded_evaluators, spans or []
        )
        evaluator_runs.extend(runs)
        evaluator_scores.extend(scores)

        # Use legacy evaluator format
        runs, scores = self._collect_results(
            sw_progress_item.eval_results,
            legacy_evaluators,
            spans or [],
        )
        evaluator_runs.extend(runs)
        evaluator_scores.extend(scores)

        # Use the appropriate spec method based on evaluation type
        if is_coded:
            spec = self._update_coded_eval_run_spec(
                evaluator_runs=evaluator_runs,
                evaluator_scores=evaluator_scores,
                eval_run_id=sw_progress_item.eval_run_id,
                execution_time=sw_progress_item.agent_execution_time,
                actual_output=sw_progress_item.agent_output,
                success=sw_progress_item.success,
                is_coded=is_coded,
            )
        else:
            spec = self._update_eval_run_spec(
                assertion_runs=evaluator_runs,
                evaluator_scores=evaluator_scores,
                eval_run_id=sw_progress_item.eval_run_id,
                execution_time=sw_progress_item.agent_execution_time,
                actual_output=sw_progress_item.agent_output,
                success=sw_progress_item.success,
                is_coded=is_coded,
            )

        await self._client.request_async(
            method=spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            headers=spec.headers,
            scoped="org" if self._is_localhost() else "tenant",
        )

    @gracefully_handle_errors
    async def update_eval_set_run(
        self,
        eval_set_run_id: str,
        evaluator_scores: dict[str, float],
        is_coded: bool = False,
        success: bool = True,
    ):
        """Update the evaluation set run status to complete."""
        spec = self._update_eval_set_run_spec(
            eval_set_run_id, evaluator_scores, is_coded, success
        )
        await self._client.request_async(
            method=spec.method,
            url=spec.endpoint,
            params=spec.params,
            json=spec.json,
            headers=spec.headers,
            scoped="org" if self._is_localhost() else "tenant",
        )

    async def handle_create_eval_set_run(self, payload: EvalSetRunCreatedEvent) -> None:
        try:
            self.evaluators = {eval.id: eval for eval in payload.evaluators}
            self.evaluator_scores = {eval.id: [] for eval in payload.evaluators}

            # Store the eval set execution ID for mapping eval runs to eval set
            self.eval_set_execution_id = payload.execution_id

            # Store the eval_set_id for this execution (needed for backend API calls)
            self.eval_set_ids[payload.execution_id] = payload.eval_set_id

            # Detect if using coded evaluators and store for this execution
            is_coded = self._is_coded_evaluator(payload.evaluators)
            self.is_coded_eval[payload.execution_id] = is_coded

            # Check if eval_set_run_id is provided (resume scenario)
            eval_set_run_id = payload.eval_set_run_id
            if eval_set_run_id:
                self.user_provided_eval_set_run_ids.add(eval_set_run_id)
                # Resume scenario: Use the provided eval_set_run_id
                # Fetch all existing eval runs from backend to populate cache
                self.is_resume_mode[payload.execution_id] = True
                logger.info(
                    f"Resume scenario: Using provided eval_set_run_id={eval_set_run_id}"
                )

                # Fetch all eval runs for this eval_set_run_id to populate cache
                # Gracefully handle errors so we don't block the resume flow
                try:
                    await self.fetch_and_cache_eval_runs(
                        eval_set_id=payload.eval_set_id,
                        eval_set_run_id=eval_set_run_id,
                        is_coded=is_coded,
                    )
                except Exception as e:
                    logger.warning(
                        f"Failed to fetch eval runs from backend during resume: {e}. "
                        "Will continue with empty cache."
                    )
            else:
                # Normal scenario: Create a new eval set run in the backend
                self.is_resume_mode[payload.execution_id] = False
                eval_set_run_id = await self.create_eval_set_run_sw(
                    eval_set_id=payload.eval_set_id,
                    agent_snapshot=self._extract_agent_snapshot(payload.entrypoint),
                    no_of_evals=payload.no_of_evals,
                    evaluators=payload.evaluators,
                    is_coded=is_coded,
                )
                logger.info(f"Created new eval_set_run_id: {eval_set_run_id}")

            self.eval_set_run_ids[payload.execution_id] = eval_set_run_id

            current_span = trace.get_current_span()
            if current_span.is_recording():
                current_span.set_attribute("eval_set_run_id", eval_set_run_id)

            # Do NOT set global trace_id override here
            # Trace IDs are set per-evaluation in handle_update_eval_run()

            logger.debug(
                f"Created eval set run with ID: {eval_set_run_id} (coded={is_coded})"
            )

        except Exception as e:
            self._format_error_message(e, "StudioWeb create eval set run error")

    async def handle_create_eval_run(self, payload: EvalRunCreatedEvent) -> None:
        try:
            logger.info(
                f"Processing CREATE_EVAL_RUN event: execution_id={payload.execution_id}, "
                f"evaluation_id={payload.eval_item.id}"
            )

            # Check if we already have an eval_run_id cached
            existing_eval_run_id = self.eval_run_ids.get(payload.execution_id)

            if existing_eval_run_id:
                # Already have eval_run_id (from previous fetch or creation)
                logger.info(
                    f"Using cached eval_run_id={existing_eval_run_id} for execution_id={payload.execution_id} "
                    f"(skipping backend fetch/create)"
                )
                return

            # Get eval_set_id, eval_set_run_id and is_coded flag
            if not self.eval_set_execution_id:
                logger.warning("Cannot process eval run: eval_set_execution_id not set")
                return

            eval_set_id = self.eval_set_ids.get(self.eval_set_execution_id)
            if not eval_set_id:
                logger.warning(
                    f"Cannot process eval run: eval_set_id not available for eval_set_execution_id={self.eval_set_execution_id}"
                )
                return

            eval_set_run_id = self.eval_set_run_ids.get(self.eval_set_execution_id)
            if not eval_set_run_id:
                logger.warning(
                    f"Cannot process eval run: eval_set_run_id not available for eval_set_execution_id={self.eval_set_execution_id}"
                )
                return

            is_coded = self.is_coded_eval.get(self.eval_set_execution_id, False)

            # Check if we're in resume mode (eval_set_run_id was provided vs created)
            is_resume = self.is_resume_mode.get(self.eval_set_execution_id, False)

            logger.info(
                f"Retrieved context: eval_set_id={eval_set_id}, eval_set_run_id={eval_set_run_id}, "
                f"is_coded={is_coded}, is_resume={is_resume}, eval_set_execution_id={self.eval_set_execution_id}"
            )

            evaluation_id = payload.eval_item.id
            eval_run_id = None

            # Only fetch from backend if we're in resume mode
            # In normal mode, we know eval runs don't exist yet (we just created eval_set_run)
            if is_resume:
                logger.info(
                    f"Resume mode: Attempting to fetch existing eval_run_id from backend for evaluation_id={evaluation_id}"
                )

                # Try to fetch existing eval run from backend (resume scenario)
                eval_run_id = await self.get_eval_run_for_evaluation(
                    eval_set_id, eval_set_run_id, evaluation_id, is_coded
                )

                if eval_run_id:
                    # Resume scenario: Found existing eval run in backend
                    self.eval_run_ids[payload.execution_id] = eval_run_id
                    logger.info(
                        f"✓ RESUME FLOW: Successfully cached eval_run_id={eval_run_id} for "
                        f"execution_id={payload.execution_id}, evaluation_id={evaluation_id}. "
                        f"Loaded from backend database."
                    )
                else:
                    logger.warning(
                        f"Resume mode but no eval_run_id found in backend for evaluation_id={evaluation_id}. "
                        f"Will create new eval run."
                    )

            # Create new eval run if not in resume mode OR if backend fetch didn't find one
            if not eval_run_id:
                logger.info(
                    f"{'Normal mode' if not is_resume else 'Resume mode (no existing run found)'}: "
                    f"Creating new eval run for evaluation_id={evaluation_id}, eval_set_run_id={eval_set_run_id}"
                )

                eval_run_id = await self.create_eval_run(
                    payload.eval_item, eval_set_run_id, is_coded
                )
                if eval_run_id:
                    # Store eval_run_id with the individual eval run's execution_id
                    self.eval_run_ids[payload.execution_id] = eval_run_id

                    logger.info(
                        f"✓ NORMAL FLOW: Successfully created and cached eval_run_id={eval_run_id} for "
                        f"execution_id={payload.execution_id}, evaluation_id={evaluation_id}, "
                        f"eval_set_run_id={eval_set_run_id} (coded={is_coded})"
                    )
                else:
                    logger.error(
                        f"✗ ERROR: create_eval_run returned None for execution_id={payload.execution_id}, "
                        f"evaluation_id={evaluation_id}, eval_set_run_id={eval_set_run_id}, is_coded={is_coded}"
                    )

        except Exception as e:
            self._format_error_message(e, "StudioWeb create eval run error")

    async def handle_update_eval_run(self, payload: EvalRunUpdatedEvent) -> None:
        try:
            logger.info(
                f"Processing UPDATE_EVAL_RUN event: execution_id={payload.execution_id}, "
                f"success={payload.success}"
            )

            eval_run_id = self.eval_run_ids.get(payload.execution_id)

            if not eval_run_id:
                logger.warning(
                    f"Cannot update eval run: eval_run_id not found in cache for "
                    f"execution_id={payload.execution_id}. Available keys: {list(self.eval_run_ids.keys())}"
                )
            else:
                logger.info(
                    f"Found eval_run_id={eval_run_id} for execution_id={payload.execution_id} in cache"
                )

            for eval_result in payload.eval_results:
                evaluator_id = eval_result.evaluator_id
                if evaluator_id in self.evaluator_scores:
                    match eval_result.result.score_type:
                        case ScoreType.NUMERICAL:
                            self.evaluator_scores[evaluator_id].append(
                                eval_result.result.score
                            )
                        case ScoreType.BOOLEAN:
                            self.evaluator_scores[evaluator_id].append(
                                100 if eval_result.result.score else 0
                            )
                        case ScoreType.ERROR:
                            self.evaluator_scores[evaluator_id].append(0)

            if eval_run_id and self.eval_set_execution_id:
                # Get the is_coded flag for this execution
                is_coded = self.is_coded_eval.get(self.eval_set_execution_id, False)

                logger.info(
                    f"Sending UPDATE to backend: eval_run_id={eval_run_id}, "
                    f"is_coded={is_coded}, success={payload.success}"
                )

                # Extract usage metrics from spans
                self._extract_usage_from_spans(payload.spans)

                await self.update_eval_run(
                    StudioWebProgressItem(
                        eval_run_id=eval_run_id,
                        eval_results=payload.eval_results,
                        success=payload.success,
                        agent_output=payload.agent_output,
                        agent_execution_time=payload.agent_execution_time,
                    ),
                    self.evaluators,
                    is_coded=is_coded,
                    spans=payload.spans,
                )

                logger.info(
                    f"✓ Successfully updated eval_run_id={eval_run_id} in backend (coded={is_coded})"
                )

        except Exception as e:
            self._format_error_message(e, "StudioWeb reporting error")

    async def handle_update_eval_set_run(self, payload: EvalSetRunUpdatedEvent) -> None:
        try:
            if eval_set_run_id := self.eval_set_run_ids.get(payload.execution_id):
                # Skip update if eval_set_run_id was provided by user
                if eval_set_run_id in self.user_provided_eval_set_run_ids:
                    logger.debug(
                        f"Skipping eval set run update for user-provided eval_set_run_id (eval_set_run_id={eval_set_run_id})"
                    )
                    return
                # Get the is_coded flag for this execution
                is_coded = self.is_coded_eval.get(payload.execution_id, False)
                await self.update_eval_set_run(
                    eval_set_run_id,
                    payload.evaluator_scores,
                    is_coded=is_coded,
                    success=payload.success,
                )
                status_str = "completed" if payload.success else "failed"
                logger.debug(
                    f"Updated eval set run with ID: {eval_set_run_id} (coded={is_coded}, status={status_str})"
                )
            else:
                logger.warning(
                    "Cannot update eval set run: eval_set_run_id not available"
                )

        except Exception as e:
            self._format_error_message(e, "StudioWeb update eval set run error")

    async def subscribe_to_eval_runtime_events(self, event_bus: EventBus) -> None:
        event_bus.subscribe(
            EvaluationEvents.CREATE_EVAL_SET_RUN, self.handle_create_eval_set_run
        )
        event_bus.subscribe(
            EvaluationEvents.CREATE_EVAL_RUN, self.handle_create_eval_run
        )
        event_bus.subscribe(
            EvaluationEvents.UPDATE_EVAL_RUN, self.handle_update_eval_run
        )
        event_bus.subscribe(
            EvaluationEvents.UPDATE_EVAL_SET_RUN, self.handle_update_eval_set_run
        )

        logger.debug("StudioWeb progress reporter subscribed to evaluation events")

    def _serialize_justification(
        self, justification: BaseModel | str | None
    ) -> str | None:
        """Serialize justification to JSON string for API compatibility.

        Args:
            justification: The justification object which could be None, a BaseModel,
                          a string, or any other JSON-serializable object

        Returns:
            JSON string representation or None if justification is None
        """
        if isinstance(justification, BaseModel):
            justification = json.dumps(justification.model_dump())

        return justification

    def _extract_agent_snapshot(self, entrypoint: str | None) -> StudioWebAgentSnapshot:
        """Extract agent snapshot from entry points configuration or low-code agent file.

        For coded agents, reads from entry-points.json configuration file.
        For low-code agents (*.json files like agent.json), reads inputSchema
        and outputSchema directly from the agent file.

        Args:
            entrypoint: The entrypoint file path to look up

        Returns:
            StudioWebAgentSnapshot with input and output schemas
        """
        if not entrypoint:
            logger.warning(
                "Entrypoint not provided - falling back to empty inputSchema "
                "and outputSchema"
            )
            return StudioWebAgentSnapshot(input_schema={}, output_schema={})

        try:
            # Check if entrypoint is a low-code agent JSON file (e.g., agent.json)
            if entrypoint.endswith(".json"):
                agent_file_path = os.path.join(os.getcwd(), entrypoint)
                if os.path.exists(agent_file_path):
                    with open(agent_file_path, "r") as f:
                        agent_data = json.load(f)

                    # Low-code agent files have inputSchema and outputSchema at root
                    input_schema = agent_data.get("inputSchema", {})
                    output_schema = agent_data.get("outputSchema", {})

                    logger.debug(
                        f"Extracted agent snapshot from low-code agent '{entrypoint}': "
                        f"inputSchema={json.dumps(input_schema)}, "
                        f"outputSchema={json.dumps(output_schema)}"
                    )

                    return StudioWebAgentSnapshot(
                        input_schema=input_schema, output_schema=output_schema
                    )

            # Fall back to entry-points.json for coded agents
            entry_points_file_path = os.path.join(
                os.getcwd(), str(UiPathConfig.entry_points_file_path)
            )
            if not os.path.exists(entry_points_file_path):
                logger.debug(
                    f"Entry points file not found at {entry_points_file_path}, "
                    "using empty schemas"
                )
                return StudioWebAgentSnapshot(input_schema={}, output_schema={})

            with open(entry_points_file_path, "r") as f:
                entry_points = json.load(f).get("entryPoints", [])

            ep = None
            for entry_point in entry_points:
                if entry_point.get("filePath") == entrypoint:
                    ep = entry_point
                    break

            if not ep:
                logger.warning(
                    f"Entrypoint {entrypoint} not found in configuration file"
                )
                return StudioWebAgentSnapshot(input_schema={}, output_schema={})

            input_schema = ep.get("input", {})
            output_schema = ep.get("output", {})

            logger.debug(
                f"Extracted agent snapshot for entrypoint '{entrypoint}': "
                f"inputSchema={json.dumps(input_schema)}, "
                f"outputSchema={json.dumps(output_schema)}"
            )

            return StudioWebAgentSnapshot(
                input_schema=input_schema, output_schema=output_schema
            )
        except Exception as e:
            logger.warning(f"Failed to extract agent snapshot: {e}")
            return StudioWebAgentSnapshot(input_schema={}, output_schema={})

    def _collect_results(
        self,
        eval_results: list[EvalItemResult],
        evaluators: dict[str, BaseLegacyEvaluator[Any]],
        spans: list[Any],
    ) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
        assertion_runs: list[dict[str, Any]] = []
        evaluator_scores_list: list[dict[str, Any]] = []

        # Extract usage metrics from spans
        usage_metrics = self._extract_usage_from_spans(spans)

        for eval_result in eval_results:
            # Skip results for evaluators not in the provided dict
            # (happens when processing mixed coded/legacy eval sets)
            if eval_result.evaluator_id not in evaluators:
                continue

            # Legacy API expects evaluatorId as GUID, convert string to GUID
            try:
                uuid.UUID(eval_result.evaluator_id)
                evaluator_id_value = eval_result.evaluator_id
            except ValueError:
                # Generate deterministic UUID5 from string
                evaluator_id_value = str(
                    uuid.uuid5(uuid.NAMESPACE_DNS, eval_result.evaluator_id)
                )

            # Convert BaseModel justification to JSON string for API compatibility
            justification = self._serialize_justification(eval_result.result.details)

            evaluator_scores_list.append(
                {
                    "type": eval_result.result.score_type.value,
                    "value": eval_result.result.score,
                    "justification": justification,
                    "evaluatorId": evaluator_id_value,
                }
            )
            assertion_runs.append(
                {
                    "status": EvaluationStatus.COMPLETED.value,
                    "evaluatorId": evaluator_id_value,
                    "completionMetrics": {
                        "duration": int(eval_result.result.evaluation_time * 1000)
                        if eval_result.result.evaluation_time
                        else 0,
                        "cost": usage_metrics["cost"],
                        "tokens": usage_metrics["tokens"] or 0,
                        "completionTokens": usage_metrics["completionTokens"] or 0,
                        "promptTokens": usage_metrics["promptTokens"] or 0,
                    },
                    "assertionSnapshot": {
                        "assertionType": evaluators[eval_result.evaluator_id].type.name,
                        "outputKey": evaluators[
                            eval_result.evaluator_id
                        ].target_output_key,
                    },
                }
            )
        return assertion_runs, evaluator_scores_list

    def _collect_coded_results(
        self,
        eval_results: list[EvalItemResult],
        evaluators: dict[str, BaseEvaluator[Any, Any, Any]],
        spans: list[Any],
    ) -> tuple[list[dict[str, Any]], list[dict[str, Any]]]:
        """Collect results for coded evaluators.

        Returns evaluatorRuns and scores in the format expected by coded eval endpoints.
        """
        evaluator_runs: list[dict[str, Any]] = []
        evaluator_scores_list: list[dict[str, Any]] = []

        # Extract usage metrics from spans
        usage_metrics = self._extract_usage_from_spans(spans)

        for eval_result in eval_results:
            # Skip results for evaluators not in the provided dict
            # (happens when processing mixed coded/legacy eval sets)
            if eval_result.evaluator_id not in evaluators:
                continue

            # Convert BaseModel justification to JSON string for API compatibility
            justification = self._serialize_justification(eval_result.result.details)

            evaluator_scores_list.append(
                {
                    "type": eval_result.result.score_type.value,
                    "value": eval_result.result.score,
                    "justification": justification,
                    "evaluatorId": eval_result.evaluator_id,
                }
            )
            evaluator_runs.append(
                {
                    "status": EvaluationStatus.COMPLETED.value,
                    "evaluatorId": eval_result.evaluator_id,
                    "result": {
                        "score": {
                            "type": eval_result.result.score_type.value,
                            "value": eval_result.result.score,
                        },
                        "justification": justification,
                    },
                    "completionMetrics": {
                        "duration": int(eval_result.result.evaluation_time * 1000)
                        if eval_result.result.evaluation_time
                        else 0,
                        "cost": usage_metrics["cost"],
                        "tokens": usage_metrics["tokens"] or 0,
                        "completionTokens": usage_metrics["completionTokens"] or 0,
                        "promptTokens": usage_metrics["promptTokens"] or 0,
                    },
                }
            )
        return evaluator_runs, evaluator_scores_list

    def _update_eval_run_spec(
        self,
        assertion_runs: list[dict[str, Any]],
        evaluator_scores: list[dict[str, Any]],
        eval_run_id: str,
        actual_output: dict[str, Any],
        execution_time: float,
        success: bool,
        is_coded: bool = False,
    ) -> RequestSpec:
        # For legacy evaluations, endpoint is without /coded
        endpoint_suffix = "coded/" if is_coded else ""

        # Determine status based on success
        status = EvaluationStatus.COMPLETED if success else EvaluationStatus.FAILED

        inner_payload: dict[str, Any] = {
            "evalRunId": eval_run_id,
            # Backend expects integer status
            "status": status.value,
            "result": {
                "output": dict(actual_output),
                "evaluatorScores": evaluator_scores,
            },
            "completionMetrics": {"duration": int(execution_time * 1000)},
            "assertionRuns": assertion_runs,
        }

        # Legacy backend expects payload wrapped in "request" field
        # Coded backend accepts payload directly
        # Both coded and legacy send payload directly at root level
        payload = inner_payload

        # Log the payload for debugging eval run updates
        agent_type = "coded" if is_coded else "low-code"
        logger.debug(
            f"Updating eval run (type={agent_type}): "
            f"evalRunId={eval_run_id}, success={success}"
        )
        logger.debug(f"Full eval run update payload: {json.dumps(payload, indent=2)}")

        return RequestSpec(
            method="PUT",
            endpoint=Endpoint(
                f"{self._get_endpoint_prefix()}execution/agents/{self._project_id}/{endpoint_suffix}evalRun"
            ),
            json=payload,
            headers=self._tenant_header(),
        )

    def _update_coded_eval_run_spec(
        self,
        evaluator_runs: list[dict[str, Any]],
        evaluator_scores: list[dict[str, Any]],
        eval_run_id: str,
        actual_output: dict[str, Any],
        execution_time: float,
        success: bool,
        is_coded: bool = False,
    ) -> RequestSpec:
        """Create update spec for coded evaluators."""
        # For coded evaluations, endpoint has /coded
        endpoint_suffix = "coded/" if is_coded else ""

        # Determine status based on success
        status = EvaluationStatus.COMPLETED if success else EvaluationStatus.FAILED

        payload: dict[str, Any] = {
            "evalRunId": eval_run_id,
            # For coded evaluations, use integer status; for legacy, use string
            "status": status.value,
            "result": {
                "output": dict(actual_output),
                "scores": evaluator_scores,
            },
            "completionMetrics": {"duration": int(execution_time * 1000)},
            "evaluatorRuns": evaluator_runs,
        }

        # Log the payload for debugging coded eval run updates
        agent_type = "coded" if is_coded else "low-code"
        logger.debug(
            f"Updating coded eval run (type={agent_type}): "
            f"evalRunId={eval_run_id}, success={success}"
        )
        logger.debug(
            f"Full coded eval run update payload: {json.dumps(payload, indent=2)}"
        )

        return RequestSpec(
            method="PUT",
            endpoint=Endpoint(
                f"{self._get_endpoint_prefix()}execution/agents/{self._project_id}/{endpoint_suffix}evalRun"
            ),
            json=payload,
            headers=self._tenant_header(),
        )

    def _create_eval_run_spec(
        self, eval_item: EvaluationItem, eval_set_run_id: str, is_coded: bool = False
    ) -> RequestSpec:
        # Legacy API expects eval IDs as GUIDs, coded accepts strings
        # Convert string IDs to deterministic GUIDs for legacy
        if is_coded:
            eval_item_id = eval_item.id
        else:
            # Try to parse as GUID, if it fails, generate deterministic GUID from string
            try:
                uuid.UUID(eval_item.id)
                eval_item_id = eval_item.id
            except ValueError:
                # Generate deterministic UUID5 from string
                eval_item_id = str(uuid.uuid5(uuid.NAMESPACE_DNS, eval_item.id))

        # Build eval snapshot based on evaluation item type
        eval_snapshot = {
            "id": eval_item_id,
            "name": eval_item.name,
            "inputs": eval_item.inputs,
        }

        # For coded evaluators, use evaluationCriterias directly
        # For legacy evaluators, extract expectedOutput from the migrated evaluationCriterias
        # (Legacy evals are migrated to EvaluationItem format with expectedOutput inside evaluationCriterias)
        if is_coded:
            eval_snapshot["evaluationCriterias"] = eval_item.evaluation_criterias
        else:
            # Legacy backend endpoint expects expectedOutput directly in evalSnapshot
            # Extract it from the first evaluator criteria (all criteria have the same expectedOutput)
            expected_output = {}
            if eval_item.evaluation_criterias:
                first_criteria = next(
                    iter(eval_item.evaluation_criterias.values()), None
                )
                if first_criteria and isinstance(first_criteria, dict):
                    expected_output = first_criteria.get("expectedOutput", {})
            eval_snapshot["expectedOutput"] = expected_output

        # For legacy evaluations, endpoint is without /coded
        endpoint_suffix = "coded/" if is_coded else ""

        inner_payload: dict[str, Any] = {
            "evalSetRunId": eval_set_run_id,
            "evalSnapshot": eval_snapshot,
            # Backend expects integer status
            "status": EvaluationStatus.IN_PROGRESS.value,
        }

        # Legacy backend expects payload wrapped in "request" field
        # Coded backend accepts payload directly
        # Both coded and legacy send payload directly at root level
        payload = inner_payload

        # Log the payload for debugging eval run reporting
        agent_type = "coded" if is_coded else "low-code"
        logger.debug(
            f"Creating eval run (type={agent_type}): "
            f"evalSetRunId={eval_set_run_id}, evalItemId={eval_item.id}"
        )
        logger.debug(f"Full eval run payload: {json.dumps(payload, indent=2)}")

        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                f"{self._get_endpoint_prefix()}execution/agents/{self._project_id}/{endpoint_suffix}evalRun"
            ),
            json=payload,
            headers=self._tenant_header(),
        )

    def _create_eval_set_run_spec(
        self,
        eval_set_id: str,
        agent_snapshot: StudioWebAgentSnapshot,
        no_of_evals: int,
        is_coded: bool = False,
    ) -> RequestSpec:
        # For legacy evaluations, endpoint is without /coded
        endpoint_suffix = "coded/" if is_coded else ""

        # Legacy API expects evalSetId as GUID, coded accepts string
        # Convert string IDs to deterministic GUIDs for legacy
        if is_coded:
            eval_set_id_value = eval_set_id
        else:
            # Try to parse as GUID, if it fails, generate deterministic GUID from string
            try:
                uuid.UUID(eval_set_id)
                eval_set_id_value = eval_set_id
            except ValueError:
                # Generate deterministic UUID5 from string
                eval_set_id_value = str(uuid.uuid5(uuid.NAMESPACE_DNS, eval_set_id))

        inner_payload: dict[str, Any] = {
            "agentId": self._project_id,
            "evalSetId": eval_set_id_value,
            "agentSnapshot": agent_snapshot.model_dump(by_alias=True),
            # Backend expects integer status
            "status": EvaluationStatus.IN_PROGRESS.value,
            "numberOfEvalsExecuted": no_of_evals,
            # Source is required by the backend (0 = coded SDK)
            "source": 0,
        }

        # Both coded and legacy send payload directly at root level
        payload = inner_payload

        # Log the payload for debugging eval set run reporting
        agent_type = "coded" if is_coded else "low-code"
        logger.info(
            f"Creating eval set run (type={agent_type}): "
            f"evalSetId={eval_set_id}, "
            f"inputSchema={json.dumps(payload.get('agentSnapshot', {}).get('inputSchema', {}))}, "
            f"outputSchema={json.dumps(payload.get('agentSnapshot', {}).get('outputSchema', {}))}"
        )
        logger.debug(f"Full eval set run payload: {json.dumps(payload, indent=2)}")

        return RequestSpec(
            method="POST",
            endpoint=Endpoint(
                f"{self._get_endpoint_prefix()}execution/agents/{self._project_id}/{endpoint_suffix}evalSetRun"
            ),
            json=payload,
            headers=self._tenant_header(),
        )

    def _update_eval_set_run_spec(
        self,
        eval_set_run_id: str,
        evaluator_scores: dict[str, float],
        is_coded: bool = False,
        success: bool = True,
    ) -> RequestSpec:
        # Legacy API expects evaluatorId as GUID, coded accepts string
        evaluator_scores_list = []
        for evaluator_id, avg_score in evaluator_scores.items():
            if is_coded:
                evaluator_id_value = evaluator_id
            else:
                # Convert string to GUID for legacy
                try:
                    uuid.UUID(evaluator_id)
                    evaluator_id_value = evaluator_id
                except ValueError:
                    # Generate deterministic UUID5 from string
                    evaluator_id_value = str(
                        uuid.uuid5(uuid.NAMESPACE_DNS, evaluator_id)
                    )

            evaluator_scores_list.append(
                {"value": avg_score, "evaluatorId": evaluator_id_value}
            )

        # For legacy evaluations, endpoint is without /coded
        endpoint_suffix = "coded/" if is_coded else ""

        # Determine status based on success
        status = EvaluationStatus.COMPLETED if success else EvaluationStatus.FAILED

        inner_payload: dict[str, Any] = {
            "evalSetRunId": eval_set_run_id,
            # Backend expects integer status
            "status": status.value,
            "evaluatorScores": evaluator_scores_list,
        }

        # Legacy backend expects payload wrapped in "request" field
        # Coded backend accepts payload directly
        # Both coded and legacy send payload directly at root level
        payload = inner_payload

        # Log the payload for debugging eval set run updates
        agent_type = "coded" if is_coded else "low-code"
        logger.info(
            f"Updating eval set run (type={agent_type}): "
            f"evalSetRunId={eval_set_run_id}, success={success}, "
            f"evaluatorScores={json.dumps(payload.get('evaluatorScores', []))}"
        )
        logger.debug(
            f"Full eval set run update payload: {json.dumps(payload, indent=2)}"
        )

        return RequestSpec(
            method="PUT",
            endpoint=Endpoint(
                f"{self._get_endpoint_prefix()}execution/agents/{self._project_id}/{endpoint_suffix}evalSetRun"
            ),
            json=payload,
            headers=self._tenant_header(),
        )

    def _get_eval_runs_spec(
        self,
        eval_set_id: str,
        eval_set_run_id: str,
        evaluation_id: str | None = None,
        is_coded: bool = False,
    ) -> RequestSpec:
        """Create request spec to GET eval runs for a given eval_set_run_id.

        Args:
            eval_set_id: The ID of the eval set
            eval_set_run_id: The ID of the eval set run
            evaluation_id: Optional evaluation ID to filter for a specific eval run
            is_coded: Whether this is a coded evaluation (vs legacy)

        Returns:
            RequestSpec for the GET request
        """
        # Build endpoint path matching backend structure:
        # Legacy: api/execution/agents/{agentId}/evalSets/{evalSetId}/evalSetRuns/{evalSetRunId}/evalRuns
        # Coded: api/execution/agents/{agentId}/coded/evalSets/{evalSetId}/evalSetRuns/{evalSetRunId}/evalRuns

        prefix = self._get_endpoint_prefix()

        if is_coded:
            endpoint_path = (
                f"{prefix}execution/agents/{self._project_id}/coded/"
                f"evalSets/{eval_set_id}/evalSetRuns/{eval_set_run_id}/evalRuns"
            )
        else:
            endpoint_path = (
                f"{prefix}execution/agents/{self._project_id}/"
                f"evalSets/{eval_set_id}/evalSetRuns/{eval_set_run_id}/evalRuns"
            )

        logger.debug(
            f"Creating GET eval runs spec: eval_set_id={eval_set_id}, "
            f"eval_set_run_id={eval_set_run_id}, evaluation_id={evaluation_id}, coded={is_coded}"
        )

        return RequestSpec(
            method="GET",
            endpoint=Endpoint(endpoint_path),
            params={},  # No query params needed - evalSetRunId is in the path
            headers=self._tenant_header(),
        )

    def _tenant_header(self) -> dict[str, str | None]:
        tenant_id = os.getenv(ENV_TENANT_ID, None)
        if not tenant_id:
            self._console.error(
                f"{ENV_TENANT_ID} env var is not set. Please run 'uipath auth'."
            )
        return {HEADER_INTERNAL_TENANT_ID: tenant_id}



import json
from pathlib import Path
from typing import Any

import click
from pydantic import ValidationError

from uipath._cli._evals._evaluator_factory import EvaluatorFactory
from uipath._cli._evals._models._evaluation_set import (
    EvaluationItem,
    EvaluationSet,
    LegacyEvaluationItem,
    LegacyEvaluationSet,
)
from uipath._cli._evals.mocks.types import InputMockingStrategy, LLMMockingStrategy
from uipath._cli._utils._console import ConsoleLogger
from uipath.eval.evaluators.base_evaluator import GenericBaseEvaluator

console = ConsoleLogger()


EVAL_SETS_DIRECTORY_NAME = "evaluations/eval-sets"


def discriminate_eval_set(data: dict[str, Any]) -> EvaluationSet | LegacyEvaluationSet:
    """Discriminate and parse evaluation set based on version field.

    Uses explicit version checking instead of Pydantic's smart union matching
    to avoid incorrect type selection when both types have matching fields.

    Args:
        data: Dictionary containing evaluation set data

    Returns:
        Either EvaluationSet (for version 1.0) or LegacyEvaluationSet
    """
    version = data.get("version")
    if isinstance(version, (int, float, str)) and float(version) >= 1:
        return EvaluationSet.model_validate(data)
    return LegacyEvaluationSet.model_validate(data)


class EvalHelpers:
    @staticmethod
    def auto_discover_eval_set() -> str:
        """Auto-discover evaluation set from {EVAL_SETS_DIRECTORY_NAME} directory.

        Returns:
            Path to the evaluation set file

        Raises:
            ValueError: If no eval set found or multiple eval sets exist
        """
        eval_sets_dir = Path(EVAL_SETS_DIRECTORY_NAME)

        if not eval_sets_dir.exists():
            raise ValueError(
                f"No '{EVAL_SETS_DIRECTORY_NAME}' directory found. "
                "Please set 'UIPATH_PROJECT_ID' env var and run 'uipath pull'."
            )

        eval_set_files = list(eval_sets_dir.glob("*.json"))

        if not eval_set_files:
            raise ValueError(
                f"No evaluation set files found in '{EVAL_SETS_DIRECTORY_NAME}' directory. "
            )

        if len(eval_set_files) > 1:
            file_names = [f.name for f in eval_set_files]
            raise ValueError(
                f"Multiple evaluation sets found: {file_names}. "
                f"Please specify which evaluation set to use: 'uipath eval [entrypoint] <eval_set_path>'"
            )

        eval_set_path = str(eval_set_files[0])
        console.info(
            f"Auto-discovered evaluation set: {click.style(eval_set_path, fg='cyan')}"
        )

        eval_set_path_obj = Path(eval_set_path)
        if not eval_set_path_obj.is_file() or eval_set_path_obj.suffix != ".json":
            raise ValueError("Evaluation set must be a JSON file")

        return eval_set_path

    @staticmethod
    def load_eval_set(
        eval_set_path: str, eval_ids: list[str] | None = None
    ) -> tuple[EvaluationSet, str]:
        """Load the evaluation set from file.

        Args:
            eval_set_path: Path to the evaluation set file
            eval_ids: Optional list of evaluation IDs to filter

        Returns:
            Tuple of (EvaluationSet, resolved_path)
        """
        # If the file doesn't exist at the given path, try looking in {EVAL_SETS_DIRECTORY_NAME}/
        resolved_path = eval_set_path
        if not Path(eval_set_path).exists():
            # Check if it's just a filename, then search in {EVAL_SETS_DIRECTORY_NAME}/
            if Path(eval_set_path).name == eval_set_path:
                eval_sets_path = Path(EVAL_SETS_DIRECTORY_NAME) / eval_set_path
                if eval_sets_path.exists():
                    resolved_path = str(eval_sets_path)

        try:
            with open(resolved_path, "r", encoding="utf-8") as f:
                data = json.load(f)
        except FileNotFoundError as e:
            raise ValueError(
                f"Evaluation set file not found: '{eval_set_path}'. "
                f"Searched in current directory and {EVAL_SETS_DIRECTORY_NAME}/ directory."
            ) from e
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Invalid JSON in evaluation set file '{resolved_path}': {str(e)}. "
                f"Please check the file for syntax errors."
            ) from e

        try:
            eval_set = discriminate_eval_set(data)
            if isinstance(eval_set, LegacyEvaluationSet):

                def migrate_evaluation_item(
                    evaluation: LegacyEvaluationItem, evaluators: list[str]
                ) -> EvaluationItem:
                    mocking_strategy = None
                    input_mocking_strategy = None
                    if (
                        evaluation.simulate_input
                        and evaluation.input_generation_instructions
                    ):
                        input_mocking_strategy = InputMockingStrategy(
                            prompt=evaluation.input_generation_instructions,
                        )
                    if evaluation.simulate_tools:
                        mocking_strategy = LLMMockingStrategy(
                            prompt=evaluation.simulation_instructions or "",
                            tools_to_simulate=evaluation.tools_to_simulate or [],
                        )
                    return EvaluationItem.model_validate(
                        {
                            "id": evaluation.id,
                            "name": evaluation.name,
                            "inputs": evaluation.inputs,
                            "expectedAgentBehavior": evaluation.expected_agent_behavior,
                            "mockingStrategy": mocking_strategy,
                            "inputMockingStrategy": input_mocking_strategy,
                            "evaluationCriterias": {
                                k: {
                                    "expectedOutput": evaluation.expected_output,
                                    "expectedAgentBehavior": evaluation.expected_agent_behavior,
                                }
                                for k in evaluators
                            },
                        }
                    )

                eval_set = EvaluationSet(
                    id=eval_set.id,
                    name=eval_set.name,
                    evaluator_refs=eval_set.evaluator_refs,
                    evaluations=[
                        migrate_evaluation_item(evaluation, eval_set.evaluator_refs)
                        for evaluation in eval_set.evaluations
                    ],
                    model_settings=eval_set.model_settings,
                )
        except ValidationError as e:
            raise ValueError(
                f"Invalid evaluation set format in '{resolved_path}': {str(e)}. "
                f"Please verify the evaluation set structure."
            ) from e
        if eval_ids:
            eval_set.extract_selected_evals(eval_ids)
        return eval_set, resolved_path

    @staticmethod
    async def load_evaluators(
        eval_set_path: str,
        evaluation_set: EvaluationSet,
        agent_model: str | None = None,
    ) -> list[GenericBaseEvaluator[Any, Any, Any]]:
        """Load evaluators referenced by the evaluation set."""
        evaluators = []
        if evaluation_set is None:
            raise ValueError("eval_set cannot be None")
        evaluators_dir = Path(eval_set_path).parent.parent / "evaluators"

        # If evaluatorConfigs is specified, use that (new field with weights)
        # Otherwise, fall back to evaluatorRefs (old field without weights)
        if (
            hasattr(evaluation_set, "evaluator_configs")
            and evaluation_set.evaluator_configs
        ):
            # Use new evaluatorConfigs field - supports weights
            evaluator_ref_ids = {ref.ref for ref in evaluation_set.evaluator_configs}
        else:
            # Fall back to old evaluatorRefs field - plain strings
            evaluator_ref_ids = set(evaluation_set.evaluator_refs)

        found_evaluator_ids = set()

        for file in evaluators_dir.glob("*.json"):
            try:
                with open(file, "r", encoding="utf-8") as f:
                    data = json.load(f)
            except json.JSONDecodeError as e:
                raise ValueError(
                    f"Invalid JSON in evaluator file '{file}': {str(e)}. "
                    f"Please check the file for syntax errors."
                ) from e

            try:
                evaluator_id = data.get("id")
                if evaluator_id in evaluator_ref_ids:
                    evaluator = EvaluatorFactory.create_evaluator(
                        data, evaluators_dir, agent_model=agent_model
                    )
                    evaluators.append(evaluator)
                    found_evaluator_ids.add(evaluator_id)
            except Exception as e:
                raise ValueError(
                    f"Failed to create evaluator from file '{file}': {str(e)}. "
                    f"Please verify the evaluator configuration."
                ) from e

        missing_evaluators = evaluator_ref_ids - found_evaluator_ids
        if missing_evaluators:
            raise ValueError(
                f"Could not find the following evaluators: {missing_evaluators}"
            )

        return evaluators



from src import ixp, du_modern, pretrained

def main():
    ixp.extract_validate()

    du_modern.extract_validate()
    du_modern.classify_extract_validate()

    pretrained.extract_validate()
    pretrained.classify_extract_validate()



from uipath.platform import UiPath
from uipath.platform.documents import ProjectType, ActionPriority

DOCUMENT_TYPE_NAME = "invoices"
FILE_PATH = "test.pdf"
ACTION_TITLE = "Test Validation Action"
ACTION_PRIORITY = ActionPriority.MEDIUM
ACTION_CATALOG = "default_du_actions"
ACTION_FOLDER = "Shared"
STORAGE_BUCKET_NAME = "du_storage_bucket"
STORAGE_BUCKET_DIRECTORY_PATH = "TestDirectory"

def extract_validate():
    uipath = UiPath()

    extraction_response = uipath.documents.extract(
        project_type=ProjectType.PRETRAINED,
        document_type_name=DOCUMENT_TYPE_NAME,
        file_path=FILE_PATH,
    )

    extraction_action = uipath.documents.create_validate_extraction_action(
        action_title=ACTION_TITLE,
        action_priority=ACTION_PRIORITY,
        action_catalog=ACTION_CATALOG,
        action_folder=ACTION_FOLDER,
        storage_bucket_name=STORAGE_BUCKET_NAME,
        storage_bucket_directory_path=STORAGE_BUCKET_DIRECTORY_PATH,
        extraction_response=extraction_response,
    )

    uipath.documents.get_validate_extraction_result(
        validation_action=extraction_action
    )

def classify_extract_validate():
    uipath = UiPath()

    classification_results = uipath.documents.classify(
        project_type=ProjectType.PRETRAINED, file_path=FILE_PATH
    )

    best_confidence_result = max(classification_results, key=lambda result: result.confidence)

    extraction_response = uipath.documents.extract(
        classification_result=best_confidence_result
    )

    extraction_action = uipath.documents.create_validate_extraction_action(
        action_title=ACTION_TITLE,
        action_priority=ACTION_PRIORITY,
        action_catalog=ACTION_CATALOG,
        action_folder=ACTION_FOLDER,
        storage_bucket_name=STORAGE_BUCKET_NAME,
        storage_bucket_directory_path=STORAGE_BUCKET_DIRECTORY_PATH,
        extraction_response=extraction_response,
    )

    uipath.documents.get_validate_extraction_result(
        validation_action=extraction_action
    )


import json

from uipath.eval.evaluators import BaseEvaluationCriteria, BaseEvaluatorConfig
from uipath.eval.evaluators.base_evaluator import BaseEvaluator
from uipath.eval.models import AgentExecution, EvaluationResult, NumericEvaluationResult
from opentelemetry.sdk.trace import ReadableSpan

class CorrectOperatorEvaluationCriteria(BaseEvaluationCriteria):
    """Evaluation criteria for the contains evaluator."""

    operator: str

class CorrectOperatorEvaluatorConfig(BaseEvaluatorConfig[CorrectOperatorEvaluationCriteria]):
    """Configuration for the contains evaluator."""

    name: str = "CorrectOperatorEvaluator"
    negated: bool = False
    default_evaluation_criteria: CorrectOperatorEvaluationCriteria = CorrectOperatorEvaluationCriteria(operator="+")

class CorrectOperatorEvaluator(BaseEvaluator[CorrectOperatorEvaluationCriteria, CorrectOperatorEvaluatorConfig, None]):
    """A custom evaluator that checks if the correct operator is being used by the agent """

    def extract_operator_from_spans(self, agent_trace: list[ReadableSpan]) -> str:
        for span in agent_trace:
            if span.name == "track_operator":
                if span.attributes:
                    input_value_as_str = span.attributes.get("input.value", "{}")
                    assert isinstance(input_value_as_str, str)
                    input_value = json.loads(input_value_as_str)
                    return input_value.get("operator")
        raise Exception(f"No 'track_operator' span found")


    @classmethod
    def get_evaluator_id(cls) -> str:
        return "CorrectOperatorEvaluator"


    async def evaluate(self, agent_execution: AgentExecution, evaluation_criteria: CorrectOperatorEvaluationCriteria) -> EvaluationResult:
        actual_operator = self.extract_operator_from_spans(agent_execution.agent_trace)
        print(actual_operator)
        is_expected_operator = evaluation_criteria.operator == actual_operator
        if self.evaluator_config.negated:
            is_expected_operator = not is_expected_operator
        return NumericEvaluationResult(
            score=float(is_expected_operator),
        )



from os import environ as env
from typing import Any

from uipath._utils.constants import (
    ENV_FOLDER_KEY,
    ENV_FOLDER_PATH,
    HEADER_FOLDER_KEY,
    HEADER_FOLDER_PATH,
)


class FolderContext:
    """Manages the folder context for UiPath automation resources.

    The FolderContext class handles information about the current folder in which
    automation resources (like processes, assets, etc.) are being accessed or modified.
    This is essential for organizing and managing resources in the UiPath Automation Cloud
    folder structure.
    """

    def __init__(self, **kwargs: Any) -> None:
        try:
            self._folder_key: str | None = env[ENV_FOLDER_KEY]
        except KeyError:
            self._folder_key = None

        try:
            self._folder_path: str | None = env[ENV_FOLDER_PATH]
        except KeyError:
            self._folder_path = None

        super().__init__(**kwargs)

    @property
    def folder_headers(self) -> dict[str, str]:
        """Get the HTTP headers for folder-based API requests.

        Returns headers containing either the folder key or folder path,
        which are used to specify the target folder for API operations.
        The folder context is essential for operations that need to be
        performed within a specific folder in UiPath Automation Cloud.

        Returns:
            dict[str, str]: A dictionary containing the appropriate folder
                header (either folder key or folder path). If no folder header is
                set as environment variable, the function returns an empty dictionary.
        """
        if self._folder_key is not None:
            return {HEADER_FOLDER_KEY: self._folder_key}
        elif self._folder_path is not None:
            return {HEADER_FOLDER_PATH: self._folder_path}
        else:
            return {}



"""Agent Models."""

from __future__ import annotations

from enum import Enum
from typing import Annotated, Any, Dict, List, Literal, Optional, Union

from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    field_validator,
    model_validator,
)
from uipath.core.guardrails import (
    BaseGuardrail,
    FieldReference,
    SpecificFieldsSelector,
    UniversalRule,
)

from uipath.platform.connections import Connection
from uipath.platform.documents import ActionPriority
from uipath.platform.guardrails import (
    BuiltInValidatorGuardrail,
)


def _decapitalize_first_letter(s: str) -> str:
    """Convert first letter to lowercase (e.g., 'SimpleText' -> 'simpleText')."""
    if not s or len(s) == 0:
        return s
    if len(s) == 1:
        return s.lower()
    return s[0].lower() + s[1:]


class AgentResourceType(str, Enum):
    """Agent resource type enumeration."""

    TOOL = "tool"
    CONTEXT = "context"
    ESCALATION = "escalation"
    MCP = "mcp"
    UNKNOWN = "unknown"  # fallback branch discriminator


class AgentToolType(str, Enum):
    """Agent tool type enumeration."""

    AGENT = "Agent"
    PROCESS = "Process"
    API = "Api"
    PROCESS_ORCHESTRATION = "ProcessOrchestration"
    INTEGRATION = "Integration"
    INTERNAL = "Internal"
    IXP = "Ixp"
    UNKNOWN = "Unknown"  # fallback branch discriminator


class AgentInternalToolType(str, Enum):
    """Agent internal tool type enumeration."""

    ANALYZE_FILES = "analyze-attachments"
    DEEP_RAG = "deep-rag"
    BATCH_TRANSFORM = "batch-transform"


class AgentEscalationRecipientType(str, Enum):
    """Agent escalation recipient type enumeration."""

    USER_ID = "UserId"
    GROUP_ID = "GroupId"
    USER_EMAIL = "UserEmail"
    ASSET_USER_EMAIL = "AssetUserEmail"
    GROUP_NAME = "GroupName"
    ASSET_GROUP_NAME = "AssetGroupName"


class AgentContextRetrievalMode(str, Enum):
    """Agent context retrieval mode enumeration."""

    SEMANTIC = "Semantic"
    STRUCTURED = "Structured"
    DEEP_RAG = "DeepRAG"
    BATCH_TRANSFORM = "BatchTransform"
    UNKNOWN = "Unknown"  # fallback branch discriminator


class AgentMessageRole(str, Enum):
    """Agent message role enumeration."""

    SYSTEM = "system"
    USER = "user"


class AgentGuardrailActionType(str, Enum):
    """Agent guardrail action type enumeration."""

    BLOCK = "block"
    ESCALATE = "escalate"
    FILTER = "filter"
    LOG = "log"
    UNKNOWN = "unknown"  # fallback branch discriminator


class AgentToolArgumentPropertiesVariant(str, Enum):
    """Agent tool argument properties variant enumeration."""

    DYNAMIC = "dynamic"
    ARGUMENT = "argument"
    STATIC = "static"
    TEXT_BUILDER = "textBuilder"


class TextTokenType(str, Enum):
    """Text token type enumeration."""

    SIMPLE_TEXT = "simpleText"
    VARIABLE = "variable"
    EXPRESSION = "expression"


class CitationMode(str, Enum):
    """Citation mode enumeration."""

    INLINE = "Inline"
    SKIP = "Skip"


class DeepRagFileExtension(str, Enum):
    """File extension enumeration for DeepRAG."""

    PDF = "pdf"
    TXT = "txt"


class BatchTransformFileExtension(str, Enum):
    """File extension enumeration for Batch Transform."""

    CSV = "csv"


class BatchTransformWebSearchGrounding(str, Enum):
    """Batch Transform web search grounding enumeration."""

    ENABLED = "Enabled"
    DISABLED = "Disabled"


class BaseCfg(BaseModel):
    """Base configuration model with common settings."""

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class ExampleCall(BaseCfg):
    """Example call for a resource containing resource I/O."""

    id: str = Field(..., alias="id")
    input: str = Field(..., alias="input")
    output: str = Field(..., alias="output")


class TextToken(BaseCfg):
    """Text token model."""

    type: TextTokenType
    raw_string: str = Field(alias="rawString")

    @field_validator("type", mode="before")
    @classmethod
    def normalize_type(cls, v: Any) -> Any:
        """Normalize type by decapitalizing first letter."""
        return _decapitalize_first_letter(v) if isinstance(v, str) else v


class BaseAgentToolArgumentProperties(BaseCfg):
    """Base tool argument properties model."""

    variant: AgentToolArgumentPropertiesVariant
    is_sensitive: bool = Field(alias="isSensitive")


class AgentToolStaticArgumentProperties(BaseAgentToolArgumentProperties):
    """Static tool argument properties model."""

    variant: Literal[AgentToolArgumentPropertiesVariant.STATIC] = Field(
        default=AgentToolArgumentPropertiesVariant.STATIC, frozen=True
    )
    value: Optional[Any]


class AgentToolArgumentArgumentProperties(BaseAgentToolArgumentProperties):
    """Agent argument argument properties model."""

    variant: Literal[AgentToolArgumentPropertiesVariant.ARGUMENT] = Field(
        default=AgentToolArgumentPropertiesVariant.ARGUMENT,
        frozen=True,
    )
    argument_path: str = Field(alias="argumentPath")


class AgentToolTextBuilderArgumentProperties(BaseAgentToolArgumentProperties):
    """Agent text builder argument properties model."""

    variant: Literal[AgentToolArgumentPropertiesVariant.TEXT_BUILDER] = Field(
        default=AgentToolArgumentPropertiesVariant.TEXT_BUILDER,
        frozen=True,
    )
    tokens: List[TextToken]


AgentToolArgumentProperties = Annotated[
    Union[
        AgentToolStaticArgumentProperties,
        AgentToolArgumentArgumentProperties,
        AgentToolTextBuilderArgumentProperties,
    ],
    Field(discriminator="variant"),
]


class BaseResourceProperties(BaseCfg):
    """Base resource properties model."""

    example_calls: Optional[list[ExampleCall]] = Field(None, alias="exampleCalls")


class AgentToolSettings(BaseCfg):
    """Agent tool settings model."""

    max_attempts: Optional[int] = Field(None, alias="maxAttempts")
    retry_delay: Optional[int] = Field(None, alias="retryDelay")
    timeout: Optional[int] = Field(None)


class BaseAgentResourceConfig(BaseCfg):
    """Base agent resource configuration model."""

    name: str
    description: str
    is_enabled: bool = Field(default=True, alias="isEnabled")
    # NOTE: this is the union discriminator; don't attach validators here.
    resource_type: Literal[
        AgentResourceType.TOOL,
        AgentResourceType.CONTEXT,
        AgentResourceType.ESCALATION,
        AgentResourceType.MCP,
        AgentResourceType.UNKNOWN,
    ] = Field(alias="$resourceType")


class AgentUnknownResourceConfig(BaseAgentResourceConfig):
    """Fallback for unknown or future resource types."""

    resource_type: Literal[AgentResourceType.UNKNOWN] = Field(
        alias="$resourceType", default=AgentResourceType.UNKNOWN, frozen=True
    )


class AgentContextQuerySetting(BaseCfg):
    """Agent context query setting model."""

    value: str | None = Field(default=None)
    description: str | None = Field(default=None)
    variant: str | None = Field(default=None)


class AgentContextValueSetting(BaseCfg):
    """Agent context value setting model."""

    value: Any = Field(...)


class DeepRagCitationModeSetting(BaseCfg):
    """DeepRAG citation mode setting model."""

    value: CitationMode = Field(...)


class DeepRagFileExtensionSetting(BaseCfg):
    """DeepRAG file extension setting model."""

    value: DeepRagFileExtension = Field(...)


class BatchTransformFileExtensionSetting(BaseCfg):
    """Batch Transform file extension setting model."""

    value: BatchTransformFileExtension = Field(...)


class BatchTransformWebSearchGroundingSetting(BaseCfg):
    """DeepRAG file extension setting model."""

    value: BatchTransformWebSearchGrounding = Field(...)


class AgentContextOutputColumn(BaseCfg):
    """Agent context output column model."""

    name: str = Field(...)
    description: Optional[str] = Field(None)


class AgentContextSettings(BaseCfg):
    """Agent context settings model."""

    result_count: int = Field(alias="resultCount")
    # Allow Unknown explicitly so we can serialize deterministically
    retrieval_mode: Literal[
        AgentContextRetrievalMode.SEMANTIC,
        AgentContextRetrievalMode.STRUCTURED,
        AgentContextRetrievalMode.DEEP_RAG,
        AgentContextRetrievalMode.BATCH_TRANSFORM,
        AgentContextRetrievalMode.UNKNOWN,
    ] = Field(alias="retrievalMode")
    threshold: float = Field(default=0)
    query: Optional[AgentContextQuerySetting] = Field(None)
    folder_path_prefix: Optional[Union[Dict[str, Any], AgentContextValueSetting]] = (
        Field(None, alias="folderPathPrefix")
    )
    file_extension: Optional[Union[Dict[str, Any], AgentContextValueSetting]] = Field(
        None, alias="fileExtension"
    )
    citation_mode: Optional[AgentContextValueSetting] = Field(
        None, alias="citationMode"
    )
    web_search_grounding: Optional[AgentContextValueSetting] = Field(
        None, alias="webSearchGrounding"
    )
    output_columns: Optional[List[AgentContextOutputColumn]] = Field(
        None, alias="outputColumns"
    )


class AgentContextResourceConfig(BaseAgentResourceConfig):
    """Agent context resource configuration model."""

    resource_type: Literal[AgentResourceType.CONTEXT] = Field(
        alias="$resourceType", default=AgentResourceType.CONTEXT, frozen=True
    )
    folder_path: str = Field(alias="folderPath")
    index_name: str = Field(alias="indexName")
    settings: AgentContextSettings = Field(..., description="Context settings")


class AgentMcpTool(BaseCfg):
    """Agent MCP tool model."""

    name: str = Field(..., alias="name")
    description: str = Field(..., alias="description")
    input_schema: Dict[str, Any] = Field(..., alias="inputSchema")
    output_schema: Optional[Dict[str, Any]] = Field(None, alias="outputSchema")
    argument_properties: Dict[str, AgentToolArgumentProperties] = Field(
        {}, alias="argumentProperties"
    )


class AgentMcpResourceConfig(BaseAgentResourceConfig):
    """Agent MCP resource configuration model."""

    resource_type: Literal[AgentResourceType.MCP] = Field(
        alias="$resourceType", default=AgentResourceType.MCP, frozen=True
    )
    folder_path: str = Field(alias="folderPath")
    slug: str = Field(..., alias="slug")
    available_tools: List[AgentMcpTool] = Field(..., alias="availableTools")


def _normalize_recipient_type(recipient: Any) -> Any:
    """Normalize recipient type from integer to enum before discrimination."""
    if not isinstance(recipient, dict):
        return recipient

    recipient_type = recipient.get("type")
    if isinstance(recipient_type, int):
        type_mapping = {
            1: AgentEscalationRecipientType.USER_ID,
            2: AgentEscalationRecipientType.GROUP_ID,
            3: AgentEscalationRecipientType.USER_EMAIL,
            4: AgentEscalationRecipientType.ASSET_USER_EMAIL,
            5: AgentEscalationRecipientType.GROUP_NAME,
            6: AgentEscalationRecipientType.ASSET_GROUP_NAME,
        }
        recipient["type"] = type_mapping.get(recipient_type, str(recipient_type))

    return recipient


class BaseEscalationRecipient(BaseCfg):
    """Base class for escalation recipients."""

    type: Union[AgentEscalationRecipientType, str] = Field(..., alias="type")


class StandardRecipient(BaseEscalationRecipient):
    """Standard recipient with value field."""

    type: Literal[
        AgentEscalationRecipientType.USER_ID,
        AgentEscalationRecipientType.GROUP_ID,
        AgentEscalationRecipientType.USER_EMAIL,
        AgentEscalationRecipientType.GROUP_NAME,
    ] = Field(..., alias="type")
    value: str = Field(..., alias="value")
    display_name: Optional[str] = Field(default=None, alias="displayName")


class AssetRecipient(BaseEscalationRecipient):
    """Asset recipient with assetName and folderPath."""

    type: Literal[
        AgentEscalationRecipientType.ASSET_USER_EMAIL,
        AgentEscalationRecipientType.ASSET_GROUP_NAME,
    ] = Field(..., alias="type")
    asset_name: str = Field(..., alias="assetName")
    folder_path: str = Field(..., alias="folderPath")


AgentEscalationRecipient = Annotated[
    Union[StandardRecipient, AssetRecipient],
    Field(discriminator="type"),
    BeforeValidator(_normalize_recipient_type),
]


class TaskTitleType(str, Enum):
    """Task title type enumeration."""

    DYNAMIC = "dynamic"
    TEXT_BUILDER = "textBuilder"


class BaseTaskTitle(BaseCfg):
    """Base class for task titles."""

    type: Union[TaskTitleType, str] = Field(..., alias="type")


class DynamicTaskTitle(BaseTaskTitle):
    """Dynamic task title with argument path."""

    type: Literal[TaskTitleType.DYNAMIC] = Field(..., alias="type")
    argument_path: str = Field(..., alias="argumentPath")


class TextBuilderTaskTitle(BaseTaskTitle):
    """Text builder task title with tokens."""

    type: Literal[TaskTitleType.TEXT_BUILDER] = Field(..., alias="type")
    tokens: List[TextToken]


TaskTitle = Annotated[
    Union[DynamicTaskTitle, TextBuilderTaskTitle],
    Field(discriminator="type"),
]


def _resolve_task_title(v: Any) -> Any:
    """Resolve taskTitleV2 and taskTitle into a single task_title field."""
    if not isinstance(v, dict):
        return v

    task_title_v2 = v.get("taskTitleV2")
    task_title = v.get("taskTitle")

    # Priority 1: Use taskTitleV2 if present
    if task_title_v2 is not None:
        title_type = (
            task_title_v2.get("type") if isinstance(task_title_v2, dict) else None
        )

        if title_type == "textBuilder":
            v["taskTitle"] = task_title_v2
        else:
            raise NotImplementedError(f"TaskTitle type '{title_type}' not implemented")

        v.pop("taskTitleV2", None)

        return v

    # Priority 2: Use taskTitle if present (legacy string support)
    if task_title is not None:
        return v

    # Priority 3: Default to "Escalation Task"
    v["taskTitle"] = "Escalation Task"

    return v


class AgentEscalationChannelProperties(BaseResourceProperties):
    """Agent escalation channel properties model."""

    app_name: str | None = Field(default=None, alias="appName")
    app_version: int = Field(..., alias="appVersion")
    folder_name: Optional[str] = Field(None, alias="folderName")
    resource_key: str | None = Field(default=None, alias="resourceKey")
    is_actionable_message_enabled: Optional[bool] = Field(
        None, alias="isActionableMessageEnabled"
    )
    actionable_message_meta_data: Optional[Any] = Field(
        None, alias="actionableMessageMetaData"
    )


class AgentEscalationChannel(BaseCfg):
    """Agent escalation channel model."""

    id: Optional[str] = Field(None, alias="id")
    name: str = Field(..., alias="name")
    type: str = Field(alias="type")
    description: str = Field(..., alias="description")
    input_schema: Dict[str, Any] = Field(..., alias="inputSchema")
    output_schema: Dict[str, Any] = Field(..., alias="outputSchema")
    argument_properties: Dict[str, AgentToolArgumentProperties] = Field(
        {}, alias="argumentProperties"
    )
    outcome_mapping: Optional[Dict[str, str]] = Field(None, alias="outcomeMapping")
    properties: AgentEscalationChannelProperties = Field(..., alias="properties")
    recipients: List[AgentEscalationRecipient] = Field(..., alias="recipients")
    task_title: Optional[Union[str, TaskTitle]] = Field(
        default="Escalation Task", alias="taskTitle"
    )
    priority: Optional[str] = None
    labels: List[str] = Field(default_factory=list)

    @model_validator(mode="before")
    @classmethod
    def _apply_task_title_resolution(cls, v: Any) -> Any:
        """Apply task title resolution."""
        return _resolve_task_title(v)


class AgentEscalationResourceConfig(BaseAgentResourceConfig):
    """Agent escalation resource configuration model."""

    id: Optional[str] = Field(None, alias="id")
    resource_type: Literal[AgentResourceType.ESCALATION] = Field(
        alias="$resourceType", default=AgentResourceType.ESCALATION, frozen=True
    )
    channels: List[AgentEscalationChannel] = Field(alias="channels")
    is_agent_memory_enabled: bool = Field(default=False, alias="isAgentMemoryEnabled")
    escalation_type: Literal[0] = Field(default=0, alias="escalationType")


class AgentIxpVsEscalationProperties(BaseCfg):
    """VS escalation properties model."""

    ixp_tool_id: str = Field(..., alias="ixpToolId")
    action_title: str | None = Field(default=None, alias="actionTitle")
    action_priority: ActionPriority | None = Field(default=None, alias="actionPriority")
    storage_bucket_name: str = Field(..., alias="storageBucketName")
    storage_bucket_folder_path: str = Field(..., alias="storageBucketFolderPath")


class AgentIxpVsEscalationResourceConfig(BaseAgentResourceConfig):
    """VS Agent escalation resource configuration model (escalationType=1)."""

    id: Optional[str] = Field(None, alias="id")
    resource_type: Literal[AgentResourceType.ESCALATION] = Field(
        alias="$resourceType", default=AgentResourceType.ESCALATION, frozen=True
    )
    channels: List[AgentEscalationChannel] = Field(alias="channels")
    is_agent_memory_enabled: bool = Field(default=False, alias="isAgentMemoryEnabled")
    escalation_type: Literal[1] = Field(default=1, alias="escalationType")
    vs_escalation_properties: AgentIxpVsEscalationProperties = Field(
        ..., alias="vsEscalationProperties"
    )


class BaseAgentToolResourceConfig(BaseAgentResourceConfig):
    """Base agent tool resource configuration model."""

    resource_type: Literal[AgentResourceType.TOOL] = Field(
        alias="$resourceType", default=AgentResourceType.TOOL, frozen=True
    )
    input_schema: Dict[str, Any] = Field(..., alias="inputSchema")


class AgentProcessToolProperties(BaseResourceProperties):
    """Agent process tool properties model."""

    folder_path: Optional[str] = Field(None, alias="folderPath")
    process_name: Optional[str] = Field(None, alias="processName")


class AgentProcessToolResourceConfig(BaseAgentToolResourceConfig):
    """Agent process tool resource configuration model."""

    type: Literal[
        AgentToolType.AGENT,
        AgentToolType.PROCESS,
        AgentToolType.API,
        AgentToolType.PROCESS_ORCHESTRATION,
    ]
    output_schema: Dict[str, Any] = Field(..., alias="outputSchema")
    properties: AgentProcessToolProperties
    settings: AgentToolSettings = Field(default_factory=AgentToolSettings)
    arguments: Dict[str, Any] = Field(default_factory=dict)
    argument_properties: Dict[str, AgentToolArgumentProperties] = Field(
        {}, alias="argumentProperties"
    )


class AgentIxpExtractionToolProperties(BaseResourceProperties):
    """Agent process tool properties model."""

    project_name: str | None = Field(None, alias="projectName")
    version_tag: str | None = Field(None, alias="versionTag")


class AgentIxpExtractionResourceConfig(BaseAgentToolResourceConfig):
    """Agent ixp extraction tool resource configuration model."""

    type: Literal[AgentToolType.IXP] = AgentToolType.IXP
    output_schema: dict[str, Any] = Field(..., alias="outputSchema")
    settings: AgentToolSettings = Field(default_factory=AgentToolSettings)
    properties: AgentIxpExtractionToolProperties


class AgentIntegrationToolParameter(BaseCfg):
    """Agent integration tool parameter model."""

    name: str = Field(..., alias="name")
    type: str = Field(..., alias="type")
    value: Optional[Any] = Field(None, alias="value")
    field_location: str = Field(..., alias="fieldLocation")

    # Optional metadata
    display_name: Optional[str] = Field(None, alias="displayName")
    display_value: Optional[str] = Field(None, alias="displayValue")
    description: Optional[str] = Field(None, alias="description")
    position: Optional[str] = Field(None, alias="position")
    field_variant: Optional[str] = Field(None, alias="fieldVariant")
    dynamic: Optional[bool] = Field(None, alias="dynamic")
    is_cascading: Optional[bool] = Field(None, alias="isCascading")
    sort_order: Optional[int] = Field(None, alias="sortOrder")
    required: Optional[bool] = Field(None, alias="required")


class AgentIntegrationToolProperties(BaseResourceProperties):
    """Agent integration tool properties model."""

    tool_path: str = Field(..., alias="toolPath")
    object_name: str = Field(..., alias="objectName")
    tool_display_name: str = Field(..., alias="toolDisplayName")
    tool_description: str = Field(..., alias="toolDescription")
    method: str = Field(..., alias="method")
    connection: Connection = Field(..., alias="connection")
    body_structure: Optional[dict[str, Any]] = Field(None, alias="bodyStructure")
    parameters: List[AgentIntegrationToolParameter] = Field(
        default_factory=list, alias="parameters"
    )


class AgentInternalAnalyzeFilesToolProperties(BaseResourceProperties):
    """Agent internal analyze files tool properties model."""

    tool_type: Literal[AgentInternalToolType.ANALYZE_FILES] = Field(
        alias="toolType", default=AgentInternalToolType.ANALYZE_FILES, frozen=True
    )


class AgentInternalDeepRagToolProperties(BaseResourceProperties):
    """Agent internal DeepRAG tool properties model."""

    tool_type: Literal[AgentInternalToolType.DEEP_RAG] = Field(
        alias="toolType", default=AgentInternalToolType.DEEP_RAG, frozen=True
    )
    settings: AgentInternalDeepRagSettings = Field(..., alias="settings")


class AgentInternalBatchTransformToolProperties(BaseResourceProperties):
    """Agent internal Batch Tranform tool properties model."""

    tool_type: Literal[AgentInternalToolType.BATCH_TRANSFORM] = Field(
        alias="toolType", default=AgentInternalToolType.BATCH_TRANSFORM, frozen=True
    )
    settings: AgentInternalBatchTransformSettings = Field(..., alias="settings")


AgentInternalToolProperties = Annotated[
    Union[
        AgentInternalAnalyzeFilesToolProperties,
        AgentInternalDeepRagToolProperties,
        AgentInternalBatchTransformToolProperties,
    ],
    Field(discriminator="tool_type"),
]


class AgentInternalDeepRagSettings(BaseCfg):
    """Agent internal DeepRAG tool settings model."""

    context_type: str = Field(..., alias="contextType")
    query: AgentContextQuerySetting = Field(...)
    folder_path_prefix: AgentContextQuerySetting | None = Field(
        default=None, alias="folderPathPrefix"
    )
    citation_mode: DeepRagCitationModeSetting = Field(..., alias="citationMode")
    file_extension: DeepRagFileExtensionSetting = Field(..., alias="fileExtension")


class AgentInternalBatchTransformSettings(BaseCfg):
    """Agent internal Batch Transform tool settings model."""

    context_type: str = Field(..., alias="contextType")
    query: AgentContextQuerySetting = Field(...)
    folder_path_prefix: AgentContextQuerySetting | None = Field(
        default=None, alias="folderPathPrefix"
    )
    file_extension: BatchTransformFileExtensionSetting = Field(
        ..., alias="fileExtension"
    )
    output_columns: List[AgentContextOutputColumn] = Field(..., alias="outputColumns")
    web_search_grounding: BatchTransformWebSearchGroundingSetting = Field(
        ..., alias="webSearchGrounding"
    )


class AgentIntegrationToolResourceConfig(BaseAgentToolResourceConfig):
    """Agent integration tool resource configuration model."""

    type: Literal[AgentToolType.INTEGRATION] = AgentToolType.INTEGRATION
    properties: AgentIntegrationToolProperties
    settings: Optional[AgentToolSettings] = Field(None)
    arguments: Optional[Dict[str, Any]] = Field(default_factory=dict)
    # is output schemas were only recently added so they will be missing in some resources
    output_schema: Optional[Dict[str, Any]] = Field(None, alias="outputSchema")


class AgentInternalToolResourceConfig(BaseAgentToolResourceConfig):
    """Agent internal tool resource configuration model."""

    type: Literal[AgentToolType.INTERNAL] = AgentToolType.INTERNAL
    properties: AgentInternalToolProperties
    settings: Optional[AgentToolSettings] = Field(None)
    arguments: Optional[Dict[str, Any]] = Field(default_factory=dict)
    output_schema: Dict[str, Any] = Field(..., alias="outputSchema")
    argument_properties: Dict[str, AgentToolArgumentProperties] = Field(
        {}, alias="argumentProperties"
    )


class AgentUnknownToolResourceConfig(BaseAgentToolResourceConfig):
    """Fallback for unknown tool types (parent normalizer sets type='Unknown')."""

    type: Literal[AgentToolType.UNKNOWN] = AgentToolType.UNKNOWN
    arguments: Optional[Dict[str, Any]] = Field(default_factory=dict)


ToolResourceConfig = Annotated[
    Union[
        AgentProcessToolResourceConfig,
        AgentIntegrationToolResourceConfig,
        AgentInternalToolResourceConfig,
        AgentIxpExtractionResourceConfig,
        AgentUnknownToolResourceConfig,  # when parent sets type="Unknown"
    ],
    Field(discriminator="type"),
]

EscalationResourceConfig = Annotated[
    Union[
        AgentEscalationResourceConfig,
        AgentIxpVsEscalationResourceConfig,
    ],
    Field(discriminator="escalation_type"),
]

AgentResourceConfig = Annotated[
    Union[
        ToolResourceConfig,  # nested discrim on 'type'
        AgentContextResourceConfig,
        EscalationResourceConfig,  # nested discrim on 'escalation_type'
        AgentMcpResourceConfig,
        AgentUnknownResourceConfig,  # when parent sets resource_type="Unknown"
    ],
    Field(discriminator="resource_type"),
]


class AgentGuardrailBlockAction(BaseModel):
    """Agent guardrail block action model."""

    action_type: Literal[AgentGuardrailActionType.BLOCK] = Field(
        alias="$actionType", default=AgentGuardrailActionType.BLOCK, frozen=True
    )
    reason: str
    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailFilterAction(BaseModel):
    """Agent guardrail filter action model."""

    action_type: Literal[AgentGuardrailActionType.FILTER] = Field(
        alias="$actionType", default=AgentGuardrailActionType.FILTER, frozen=True
    )
    fields: List[FieldReference]
    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailSeverityLevel(str, Enum):
    """Severity level enumeration."""

    ERROR = "Error"
    INFO = "Info"
    WARNING = "Warning"


class AgentGuardrailLogAction(BaseModel):
    """Agent guardrail log action model."""

    action_type: Literal[AgentGuardrailActionType.LOG] = Field(
        alias="$actionType", default=AgentGuardrailActionType.LOG, frozen=True
    )
    message: Optional[str] = Field(None, alias="message")
    severity_level: AgentGuardrailSeverityLevel = Field(alias="severityLevel")
    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailEscalateActionApp(BaseModel):
    """Agent guardrail escalate action app model."""

    id: Optional[str] = None
    version: int
    name: str
    folder_id: Optional[str] = Field(None, alias="folderId")
    folder_name: str = Field(alias="folderName")
    app_process_key: Optional[str] = Field(None, alias="appProcessKey")
    runtime: Optional[str] = None
    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailEscalateAction(BaseModel):
    """Agent guardrail escalate action model."""

    action_type: Literal[AgentGuardrailActionType.ESCALATE] = Field(
        alias="$actionType", default=AgentGuardrailActionType.ESCALATE, frozen=True
    )
    app: AgentGuardrailEscalateActionApp
    recipient: "AgentEscalationRecipient"  # forward ref ok
    model_config = ConfigDict(populate_by_name=True, extra="allow")


class AgentGuardrailUnknownAction(BaseModel):
    """Fallback for unknown guardrail actions."""

    action_type: Literal[AgentGuardrailActionType.UNKNOWN] = Field(
        alias="$actionType", default=AgentGuardrailActionType.UNKNOWN, frozen=True
    )
    # Accept arbitrary payload for forward-compat
    details: Optional[Dict[str, Any]] = None
    model_config = ConfigDict(populate_by_name=True, extra="allow")


GuardrailAction = Annotated[
    Union[
        AgentGuardrailBlockAction,
        AgentGuardrailFilterAction,
        AgentGuardrailLogAction,
        AgentGuardrailEscalateAction,
        AgentGuardrailUnknownAction,  # when parent sets $actionType="unknown"
    ],
    Field(discriminator="action_type"),
]


class AgentBuiltInValidatorGuardrail(BuiltInValidatorGuardrail):
    """Agent built-in validator guardrail with action capabilities."""

    action: GuardrailAction = Field(
        ..., description="Action to take when guardrail is triggered"
    )

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentWordOperator(str, Enum):
    """Word operator enumeration."""

    CONTAINS = "contains"
    DOES_NOT_CONTAIN = "doesNotContain"
    DOES_NOT_END_WITH = "doesNotEndWith"
    DOES_NOT_EQUAL = "doesNotEqual"
    DOES_NOT_START_WITH = "doesNotStartWith"
    ENDS_WITH = "endsWith"
    EQUALS = "equals"
    IS_EMPTY = "isEmpty"
    IS_NOT_EMPTY = "isNotEmpty"
    MATCHES_REGEX = "matchesRegex"
    STARTS_WITH = "startsWith"


class AgentWordRule(BaseModel):
    """Word rule model."""

    rule_type: Literal["word"] = Field(alias="$ruleType")
    field_selector: AgentFieldSelector = Field(alias="fieldSelector")
    operator: AgentWordOperator
    value: str | None = None

    model_config = ConfigDict(populate_by_name=True, extra="allow")

    @field_validator("operator", mode="before")
    @classmethod
    def normalize_operator(cls, v: Any) -> Any:
        """Normalize operator by decapitalizing first letter."""
        return _decapitalize_first_letter(v) if isinstance(v, str) else v


class AgentAllFieldsSelector(BaseModel):
    """All fields selector."""

    selector_type: Literal["all"] = Field(alias="$selectorType")

    model_config = ConfigDict(populate_by_name=True, extra="allow")


AgentFieldSelector = Annotated[
    AgentAllFieldsSelector | SpecificFieldsSelector,
    Field(discriminator="selector_type"),
]


class AgentNumberOperator(str, Enum):
    """Number operator enumeration."""

    DOES_NOT_EQUAL = "doesNotEqual"
    EQUALS = "equals"
    GREATER_THAN = "greaterThan"
    GREATER_THAN_OR_EQUAL = "greaterThanOrEqual"
    LESS_THAN = "lessThan"
    LESS_THAN_OR_EQUAL = "lessThanOrEqual"


class AgentNumberRule(BaseModel):
    """Number rule model."""

    rule_type: Literal["number"] = Field(alias="$ruleType")
    field_selector: AgentFieldSelector = Field(alias="fieldSelector")
    operator: AgentNumberOperator
    value: float

    model_config = ConfigDict(populate_by_name=True, extra="allow")

    @field_validator("operator", mode="before")
    @classmethod
    def normalize_operator(cls, v: Any) -> Any:
        """Normalize operator by decapitalizing first letter."""
        return _decapitalize_first_letter(v) if isinstance(v, str) else v


class AgentBooleanOperator(str, Enum):
    """Boolean operator enumeration."""

    EQUALS = "equals"


class AgentBooleanRule(BaseModel):
    """Boolean rule model."""

    rule_type: Literal["boolean"] = Field(alias="$ruleType")
    field_selector: AgentFieldSelector = Field(alias="fieldSelector")
    operator: AgentBooleanOperator
    value: bool

    model_config = ConfigDict(populate_by_name=True, extra="allow")

    @field_validator("operator", mode="before")
    @classmethod
    def normalize_operator(cls, v: Any) -> Any:
        """Normalize operator by decapitalizing first letter."""
        return _decapitalize_first_letter(v) if isinstance(v, str) else v


AgentRule = Annotated[
    AgentWordRule | AgentNumberRule | AgentBooleanRule | UniversalRule,
    Field(discriminator="rule_type"),
]


class AgentCustomGuardrail(BaseGuardrail):
    """Agent custom guardrail with action capabilities."""

    guardrail_type: Literal["custom"] = Field(alias="$guardrailType")
    rules: list[AgentRule]

    action: GuardrailAction = Field(
        ..., description="Action to take when guardrail is triggered"
    )

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )


class AgentUnknownGuardrail(BaseCfg):
    """Fallback wrapper for unknown guardrail kinds."""

    guardrail_type: Literal["unknown"] = Field(
        alias="$guardrailType", default="unknown", frozen=True
    )
    # store the original payload under 'raw' for round-trip/debug
    raw: Dict[str, Any]


AgentGuardrail = Annotated[
    Union[
        # known kinds from SDK
        AgentCustomGuardrail,
        AgentBuiltInValidatorGuardrail,
        # unknown kind fallback
        AgentUnknownGuardrail,
    ],
    Field(discriminator="guardrail_type"),
]


class AgentMetadata(BaseCfg):
    """Agent metadata model."""

    is_conversational: bool = Field(alias="isConversational")
    storage_version: str = Field(alias="storageVersion")


class AgentMessage(BaseCfg):
    """Agent message model."""

    role: Literal[AgentMessageRole.SYSTEM, AgentMessageRole.USER]
    content: str
    content_tokens: Optional[List[TextToken]] = Field(None, alias="contentTokens")

    @field_validator("role", mode="before")
    @classmethod
    def normalize_role(cls, v: Any) -> Any:
        """Normalize role to lowercase enum/string."""
        return v.lower() if isinstance(v, str) else v


class AgentByomProperties(BaseCfg):
    """Agent byom properties model."""

    connection_id: str = Field(alias="connectionId")
    connector_key: str = Field(alias="connectorKey")


class AgentSettings(BaseCfg):
    """Agent settings model."""

    engine: str
    model: str
    max_tokens: int = Field(..., alias="maxTokens")
    temperature: float
    byom_properties: Optional[AgentByomProperties] = Field(None, alias="byomProperties")
    max_iterations: Optional[int] = Field(None, alias="maxIterations")


class AgentDefinition(BaseModel):
    """Unified agent definition with parent-level normalization for guardrails and resources."""

    input_schema: Dict[str, Any] = Field(..., alias="inputSchema")
    output_schema: Dict[str, Any] = Field(..., alias="outputSchema")
    guardrails: Optional[List[AgentGuardrail]] = Field(None)

    id: Optional[str] = None
    name: Optional[str] = None
    metadata: Optional[AgentMetadata] = None
    messages: List[AgentMessage]

    version: str = "1.0.0"
    resources: List[AgentResourceConfig] = Field(default_factory=list)
    features: List[Any] = Field(default_factory=list)
    settings: AgentSettings

    model_config = ConfigDict(
        validate_by_name=True, validate_by_alias=True, extra="allow"
    )

    @property
    def is_conversational(self) -> bool:
        """Checks the settings.engine property to determine if the agent is conversational."""
        if hasattr(self, "metadata") and self.metadata:
            metadata = self.metadata
            if hasattr(metadata, "is_conversational"):
                return metadata.is_conversational
        return False

    @staticmethod
    def _normalize_guardrails(v: Dict[str, Any]) -> None:
        guards = v.get("guardrails")
        if not isinstance(guards, list):
            return

        normalized = []
        for g in guards:
            if not isinstance(g, dict):
                normalized.append(g)
                continue

            gt = g.get("$guardrailType") or g.get("guardrail_type")
            # Normalize to the expected discriminator values
            if isinstance(gt, str):
                gt_lower = gt.lower()
                if gt_lower == "custom":
                    g["$guardrailType"] = "custom"
                elif gt_lower == "builtinvalidator":
                    g["$guardrailType"] = "builtInValidator"
                else:
                    # Unknown guardrail type
                    normalized.append({"guardrail_type": "unknown", "raw": g})
                    continue
            else:
                # Non-string guardrail type
                normalized.append({"guardrail_type": "unknown", "raw": g})
                continue

            # Normalize the action if present
            action = g.get("action")
            if isinstance(action, dict):
                at = action.get("$actionType")
                if isinstance(at, str):
                    at_lower = at.lower()
                    if at_lower in {"block", "filter", "log", "escalate"}:
                        # Valid action type, keep as-is or normalize case if needed
                        g["action"]["$actionType"] = at_lower
                    else:
                        # Unknown action type
                        g["action"] = {"$actionType": "unknown", "details": action}
                else:
                    # Non-string action type
                    g["action"] = {"$actionType": "unknown", "details": action}

            normalized.append(g)

        v["guardrails"] = normalized

    @staticmethod
    def _normalize_resources(v: Dict[str, Any]) -> None:
        KNOWN_RES = {"tool", "context", "escalation", "mcp"}
        TOOL_MAP = {
            "agent": "Agent",
            "process": "Process",
            "api": "Api",
            "processorchestration": "ProcessOrchestration",
            "integration": "Integration",
            "internal": "Internal",
            "ixp": "Ixp",
            "unknown": "Unknown",
        }
        CONTEXT_MODE_MAP = {
            "semantic": "Semantic",
            "structured": "Structured",
            "deeprag": "DeepRAG",
            "batchtransform": "BatchTransform",
            "unknown": "Unknown",
        }

        res_list = v.get("resources")
        if not isinstance(res_list, list):
            return

        out = []
        for res in res_list:
            if not isinstance(res, dict):
                out.append(res)
                continue

            rt = res.get("$resourceType") or res.get("resource_type")
            res["$resourceType"] = (
                rt.lower()
                if isinstance(rt, str) and rt.lower() in KNOWN_RES
                else "unknown"
            )

            if res["$resourceType"] == "tool":
                t = res.get("type")
                res["type"] = (
                    TOOL_MAP.get(t.lower(), "Unknown")
                    if isinstance(t, str)
                    else "Unknown"
                )

            if res["$resourceType"] == "context":
                settings = res.get("settings", {})
                rm = settings.get("retrievalMode") or settings.get("retrieval_mode")
                settings["retrievalMode"] = (
                    CONTEXT_MODE_MAP.get(rm.lower(), "Unknown")
                    if isinstance(rm, str)
                    else "Unknown"
                )
                res["settings"] = settings

            out.append(res)

        v["resources"] = out

    @model_validator(mode="before")
    @classmethod
    def _normalize_all(cls, v: Any) -> Any:
        if not isinstance(v, dict):
            return v
        cls._normalize_guardrails(v)
        cls._normalize_resources(v)
        return v


LowCodeAgentDefinition = AgentDefinition



import json
import os
import uuid
from datetime import date, datetime, time
from pathlib import Path
from typing import Literal
from urllib.parse import urlparse

import click

from uipath.platform.common import UiPathConfig

from ..._utils._bindings import ResourceOverwrite, ResourceOverwriteParser
from ..._utils.constants import ENV_UIPATH_ACCESS_TOKEN
from ..spinner import Spinner
from ._console import ConsoleLogger
from ._studio_project import (
    NonCodedAgentProjectException,
    StudioClient,
    StudioProjectMetadata,
)


def get_claim_from_token(claim_name: str) -> str | None:
    import jwt

    token = os.getenv(ENV_UIPATH_ACCESS_TOKEN)
    if not token:
        raise Exception("JWT token not available")
    decoded_token = jwt.decode(token, options={"verify_signature": False})
    return decoded_token.get(claim_name)


def environment_options(function):
    function = click.option(
        "--alpha",
        "environment",
        flag_value="alpha",
        help="Use alpha environment",
    )(function)
    function = click.option(
        "--staging",
        "environment",
        flag_value="staging",
        help="Use staging environment",
    )(function)
    function = click.option(
        "--cloud",
        "environment",
        flag_value="cloud",
        help="Use production environment",
    )(function)
    return function


def get_env_vars(spinner: Spinner | None = None) -> list[str]:
    base_url = os.environ.get("UIPATH_URL")
    token = os.environ.get("UIPATH_ACCESS_TOKEN")

    if not all([base_url, token]):
        if spinner:
            spinner.stop()
        click.echo(
            "❌ Missing required environment variables. Please check your .env file contains:"
        )
        click.echo("UIPATH_URL, UIPATH_ACCESS_TOKEN")
        click.get_current_context().exit(1)

    assert base_url and token
    return [base_url, token]


def serialize_object(obj):
    """Recursively serializes an object and all its nested components."""
    # Handle Pydantic models
    if hasattr(obj, "model_dump"):
        return serialize_object(obj.model_dump(by_alias=True))
    elif hasattr(obj, "dict"):
        return serialize_object(obj.dict())
    elif hasattr(obj, "to_dict"):
        return serialize_object(obj.to_dict())
    elif isinstance(obj, (datetime, date, time)):
        return obj.isoformat()
    # Handle dictionaries
    elif isinstance(obj, dict):
        return {k: serialize_object(v) for k, v in obj.items()}
    # Handle lists
    elif isinstance(obj, list):
        return [serialize_object(item) for item in obj]
    # Handle other iterable objects (convert to dict first)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes)):
        try:
            return serialize_object(dict(obj))
        except (TypeError, ValueError):
            return obj
    # UUIDs must be serialized explicitly
    elif isinstance(obj, uuid.UUID):
        return str(obj)
    # Return primitive types as is
    else:
        return obj


def get_org_scoped_url(base_url: str) -> str:
    """Get organization scoped URL from base URL.

    Args:
        base_url: The base URL to scope

    Returns:
        str: The organization scoped URL
    """
    parsed = urlparse(base_url)
    org_name, *_ = parsed.path.strip("/").split("/")
    org_scoped_url = f"{parsed.scheme}://{parsed.netloc}/{org_name}"
    return org_scoped_url


def clean_directory(directory: str) -> None:
    """Clean up Python files in the specified directory.

    Args:
        directory (str): Path to the directory to clean.

    This function removes all Python files (*.py) from the specified directory.
    It's used to prepare a directory for a quickstart agent/coded MCP server.
    """
    for file_name in os.listdir(directory):
        file_path = os.path.join(directory, file_name)

        if os.path.isfile(file_path) and file_name.endswith(".py"):
            os.remove(file_path)


async def ensure_coded_agent_project(studio_client: StudioClient):
    try:
        await studio_client.ensure_coded_agent_project_async()
    except NonCodedAgentProjectException:
        console = ConsoleLogger()
        console.error(
            "The targeted Studio Web project is not of type coded agent. Please check the UIPATH_PROJECT_ID environment variable."
        )


async def may_override_files(
    studio_client: StudioClient, scope: Literal["remote", "local"]
) -> bool:
    from datetime import datetime

    from packaging import version

    remote_metadata = await studio_client.get_project_metadata_async()
    if not remote_metadata:
        return True

    metadata_file_path = UiPathConfig.studio_metadata_file_path

    local_code_version = None

    if os.path.isfile(metadata_file_path):
        with open(metadata_file_path, "r") as f:
            local_data = json.load(f)
            local_metadata = StudioProjectMetadata.model_validate(local_data)
            local_code_version = local_metadata.code_version
        if version.parse(local_code_version) >= version.parse(
            remote_metadata.code_version
        ):
            return True

    local_version_display = local_code_version if local_code_version else "Not Set"

    try:
        push_date = datetime.fromisoformat(remote_metadata.last_push_date)
        formatted_date = push_date.strftime("%b %d, %Y at %I:%M %p UTC")
    except (ValueError, TypeError):
        formatted_date = remote_metadata.last_push_date

    console = ConsoleLogger()
    console.warning("Your local version is behind the remote version.")
    console.info(f"  Remote version:  {remote_metadata.code_version}")
    console.info(f"  Local version:   {local_version_display}")
    console.info(f"  Last publisher:  {remote_metadata.last_push_author}")
    console.info(f"  Last push date:  {formatted_date}")

    return console.confirm(
        f"Do you want to proceed with overwriting the {scope} files?"
    )


async def read_resource_overwrites_from_file(
    directory_path: str | None = None,
) -> dict[str, ResourceOverwrite]:
    """Read resource overwrites from a JSON file."""
    config_file_name = UiPathConfig.config_file_name
    if directory_path is not None:
        file_path = Path(f"{directory_path}/{config_file_name}")
    else:
        file_path = Path(f"{config_file_name}")

    overwrites_dict = {}

    try:
        with open(file_path, "r") as f:
            data = json.load(f)
            resource_overwrites = (
                data.get("runtime", {})
                .get("internalArguments", {})
                .get("resourceOverwrites", {})
            )
            for key, value in resource_overwrites.items():
                overwrites_dict[key] = ResourceOverwriteParser.parse(key, value)

    # Return empty dict if file doesn't exist or invalid json
    except FileNotFoundError:
        pass
    except json.JSONDecodeError:
        pass

    return overwrites_dict



"""UiPath Functions Runtime Factory - discovery and creation of function-based runtimes."""

import json
import logging
from pathlib import Path
from typing import Any

from uipath.runtime import (
    UiPathRuntimeFactorySettings,
    UiPathRuntimeProtocol,
    UiPathRuntimeStorageProtocol,
)

from .debug import UiPathDebugFunctionsRuntime
from .runtime import UiPathFunctionsRuntime

logger = logging.getLogger(__name__)


class UiPathFunctionsRuntimeFactory:
    """Factory for discovering and creating function-based runtimes."""

    def __init__(self, config_path: str = "uipath.json", base_dir: str | None = None):
        """Initialize the factory with the path to uipath.json configuration."""
        self.config_path = Path(config_path)
        self.base_dir = Path(base_dir) if base_dir else self.config_path.parent
        self._config: dict[str, Any] | None = None

    def _load_config(self) -> dict[str, Any]:
        """Load uipath.json configuration with caching."""
        if self._config is not None:
            return self._config

        if not self.config_path.exists():
            logger.warning(f"Config file not found: {self.config_path}")
            return {}

        try:
            with open(self.config_path) as f:
                self._config = json.load(f)
        except json.JSONDecodeError as e:
            logger.error(f"Invalid JSON in {self.config_path}: {e}")
            return {}

        return self._config

    def discover_entrypoints(self) -> list[str]:
        """Discover all function entrypoints from uipath.json."""
        config = self._load_config()
        return list(config.get("functions", {}).keys())

    async def get_storage(self) -> UiPathRuntimeStorageProtocol | None:
        """Get storage protocol if any (placeholder for protocol compliance)."""
        return None

    async def get_settings(self) -> UiPathRuntimeFactorySettings | None:
        """Get factory settings for coded functions.

        Coded functions don't need span filtering - all spans are relevant
        since developers have full control over instrumentation.

        Low-code agents (LangGraph) need filtering due to framework overhead.
        """
        return None

    async def new_runtime(
        self, entrypoint: str, runtime_id: str, **kwargs
    ) -> UiPathRuntimeProtocol:
        """Create a new runtime instance for the given entrypoint."""
        return self._create_runtime(entrypoint)

    async def dispose(self) -> None:
        """Dispose resources if any (placeholder for interface compliance)."""
        pass

    def _create_runtime(self, entrypoint: str) -> UiPathRuntimeProtocol:
        """Create runtime instance from entrypoint specification."""
        config = self._load_config()
        functions = config.get("functions", {})

        if entrypoint not in functions:
            raise ValueError(
                f"Entrypoint '{entrypoint}' not found in uipath.json. "
                f"Available: {', '.join(functions.keys())}"
            )

        func_spec = functions[entrypoint]

        if ":" not in func_spec:
            raise ValueError(
                f"Invalid function specification: '{func_spec}'. "
                "Expected format: 'path/to/file.py:function_name'"
            )

        file_path, function_name = func_spec.rsplit(":", 1)
        full_path = self.base_dir / file_path

        if not full_path.exists():
            raise ValueError(f"File not found: {full_path}")

        inner = UiPathFunctionsRuntime(str(full_path), function_name, entrypoint)
        return UiPathDebugFunctionsRuntime(
            delegate=inner,
            entrypoint_path=str(full_path),
        )



"""Init file for resume triggers module."""

from ._enums import PropertyName, TriggerMarker
from ._protocol import (
    UiPathResumeTriggerCreator,
    UiPathResumeTriggerHandler,
    UiPathResumeTriggerReader,
)

__all__ = [
    "UiPathResumeTriggerReader",
    "UiPathResumeTriggerCreator",
    "UiPathResumeTriggerHandler",
    "PropertyName",
    "TriggerMarker",
]



"""Models for MCP Servers in UiPath Orchestrator."""

from datetime import datetime
from enum import IntEnum
from typing import Optional

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class McpServerType(IntEnum):
    """Enumeration of MCP server types."""

    UiPath = 0  # Processes, Agents, Activities
    Command = 1  # npx, uvx
    Coded = 2  # PackageType.McpServer
    SelfHosted = 3  # tunnel to (externally) self-hosted server
    Remote = 4  # HTTP connection to remote MCP server
    ProcessAssistant = 5  # Dynamic user process assistant


class McpServerStatus(IntEnum):
    """Enumeration of MCP server statuses."""

    Disconnected = 0
    Connected = 1


class McpServer(BaseModel):
    """Model representing an MCP server in UiPath Orchestrator."""

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        use_enum_values=True,
        arbitrary_types_allowed=True,
        extra="allow",
    )

    id: Optional[str] = None
    name: Optional[str] = None
    slug: Optional[str] = None
    description: Optional[str] = None
    version: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    is_active: Optional[bool] = None
    type: Optional[McpServerType] = None
    status: Optional[McpServerStatus] = None
    command: Optional[str] = None
    arguments: Optional[str] = None
    environment_variables: Optional[str] = None
    process_key: Optional[str] = None
    folder_key: Optional[str] = None
    runtimes_count: Optional[int] = None
    mcp_url: Optional[str] = None



#!/usr/bin/env python3
"""Custom linter to check for httpx.Client() usage.

This script checks for direct usage of httpx.Client() without using the
get_httpx_client_kwargs() function, which is required for proper SSL
and proxy configuration in the UiPath Python SDK.
"""

import ast
import sys
from pathlib import Path
from typing import NamedTuple


class LintViolation(NamedTuple):
    """Represents a linting violation."""

    filename: str
    line: int
    column: int
    message: str
    rule_code: str


class HttpxClientChecker(ast.NodeVisitor):
    """AST visitor to check for httpx.Client() usage violations."""

    def __init__(self, filename: str):
        """Initialize the checker with a filename.

        Args:
            filename: The path to the file being checked.
        """
        self.filename = filename
        self.violations: list[LintViolation] = []
        self.has_httpx_import = False
        self.has_get_httpx_client_kwargs_import = False
        # Track variables that contain get_httpx_client_kwargs
        self.variables_with_httpx_kwargs: set[str] = set()

    def visit_Import(self, node: ast.Import) -> None:
        """Check for httpx imports."""
        for alias in node.names:
            if alias.name == "httpx":
                self.has_httpx_import = True
        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        """Check for imports from httpx or get_httpx_client_kwargs."""
        if node.module == "httpx":
            self.has_httpx_import = True
        elif node.module and "get_httpx_client_kwargs" in [
            alias.name for alias in (node.names or [])
        ]:
            self.has_get_httpx_client_kwargs_import = True
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        """Track variable assignments that use get_httpx_client_kwargs."""
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            var_name = node.targets[0].id
            if self._assignment_uses_get_httpx_client_kwargs(node.value):
                self.variables_with_httpx_kwargs.add(var_name)
        self.generic_visit(node)

    def _assignment_uses_get_httpx_client_kwargs(self, value_node: ast.AST) -> bool:
        """Check if an assignment value uses get_httpx_client_kwargs."""
        if isinstance(value_node, ast.Call):
            # Direct call: var = get_httpx_client_kwargs()
            if isinstance(value_node.func, ast.Name):
                if value_node.func.id == "get_httpx_client_kwargs":
                    return True
            elif isinstance(value_node.func, ast.Attribute):
                if value_node.func.attr == "get_httpx_client_kwargs":
                    return True
        elif isinstance(value_node, ast.Dict):
            # Dictionary that spreads get_httpx_client_kwargs: {..., **get_httpx_client_kwargs()}
            for key in value_node.keys:
                if key is None:  # This is a **kwargs expansion
                    # Find corresponding value
                    idx = value_node.keys.index(key)
                    if idx < len(value_node.values):
                        spread_value = value_node.values[idx]
                        if isinstance(spread_value, ast.Call):
                            if isinstance(spread_value.func, ast.Name):
                                if spread_value.func.id == "get_httpx_client_kwargs":
                                    return True
                            elif isinstance(spread_value.func, ast.Attribute):
                                if spread_value.func.attr == "get_httpx_client_kwargs":
                                    return True
                        elif isinstance(spread_value, ast.Name):
                            # Spreading another variable that might contain httpx kwargs
                            if spread_value.id in self.variables_with_httpx_kwargs:
                                return True
        return False

    def visit_Call(self, node: ast.Call) -> None:
        """Check for httpx.Client() and httpx.AsyncClient() calls."""
        if self._is_httpx_client_call(node):
            # Check if this is a proper usage with get_httpx_client_kwargs
            if not self._is_using_get_httpx_client_kwargs(node):
                client_type = self._get_client_type(node)
                violation = LintViolation(
                    filename=self.filename,
                    line=node.lineno,
                    column=node.col_offset,
                    message=f"Use **get_httpx_client_kwargs() with {client_type}() - should be: {client_type}(**get_httpx_client_kwargs())",
                    rule_code="UIPATH001",
                )
                self.violations.append(violation)

        self.generic_visit(node)

    def _is_httpx_client_call(self, node: ast.Call) -> bool:
        """Check if the call is httpx.Client() or httpx.AsyncClient()."""
        if isinstance(node.func, ast.Attribute):
            if (
                isinstance(node.func.value, ast.Name)
                and node.func.value.id == "httpx"
                and node.func.attr in ("Client", "AsyncClient")
            ):
                return True
        elif isinstance(node.func, ast.Name) and node.func.id in (
            "Client",
            "AsyncClient",
        ):
            # This could be a direct Client/AsyncClient import, check if httpx is imported
            return self.has_httpx_import
        return False

    def _get_client_type(self, node: ast.Call) -> str:
        """Get the client type name (Client or AsyncClient)."""
        if isinstance(node.func, ast.Attribute):
            return f"httpx.{node.func.attr}"
        elif isinstance(node.func, ast.Name):
            return node.func.id
        return "httpx.Client"

    def _is_using_get_httpx_client_kwargs(self, node: ast.Call) -> bool:
        """Check if the httpx.Client() call is using **get_httpx_client_kwargs()."""
        # Check if there are any **kwargs that use get_httpx_client_kwargs directly
        for keyword in node.keywords:
            if keyword.arg is None:  # This is a **kwargs expansion
                if isinstance(keyword.value, ast.Call):
                    if isinstance(keyword.value.func, ast.Name):
                        if keyword.value.func.id == "get_httpx_client_kwargs":
                            return True
                    elif isinstance(keyword.value.func, ast.Attribute):
                        if keyword.value.func.attr == "get_httpx_client_kwargs":
                            return True
                elif isinstance(keyword.value, ast.Name):
                    # Check if this variable might contain get_httpx_client_kwargs
                    # This handles cases like: **client_kwargs where client_kwargs was built from get_httpx_client_kwargs
                    var_name = keyword.value.id
                    if self._variable_contains_get_httpx_client_kwargs(var_name):
                        return True

        # Also check if it's the ONLY argument and it's **get_httpx_client_kwargs()
        # This handles cases like: httpx.Client(**get_httpx_client_kwargs())
        if len(node.args) == 0 and len(node.keywords) == 1:
            keyword = node.keywords[0]
            if keyword.arg is None and isinstance(keyword.value, ast.Call):
                if isinstance(keyword.value.func, ast.Name):
                    if keyword.value.func.id == "get_httpx_client_kwargs":
                        return True
                elif isinstance(keyword.value.func, ast.Attribute):
                    if keyword.value.func.attr == "get_httpx_client_kwargs":
                        return True

        return False

    def _variable_contains_get_httpx_client_kwargs(self, var_name: str) -> bool:
        """Check if a variable was built using get_httpx_client_kwargs()."""
        # Check if we've tracked this variable as containing httpx kwargs
        if var_name in self.variables_with_httpx_kwargs:
            return True

        # Fallback: Simple heuristic based on naming patterns
        # This handles cases where the variable assignment might be complex
        if "client_kwargs" in var_name.lower() or "httpx_kwargs" in var_name.lower():
            return True

        return False


def check_file(filepath: Path) -> list[LintViolation]:
    """Check a single Python file for httpx.Client() violations."""
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            content = f.read()

        tree = ast.parse(content, filename=str(filepath))
        checker = HttpxClientChecker(str(filepath))
        checker.visit(tree)
        return checker.violations

    except SyntaxError:
        # Skip files with syntax errors
        return []
    except Exception as e:
        print(f"Error checking {filepath}: {e}", file=sys.stderr)
        return []


def main():
    """Main function to run the linter."""
    if len(sys.argv) > 1:
        paths = [Path(p) for p in sys.argv[1:]]
    else:
        # Default to checking src and tests directories
        paths = [Path("src"), Path("tests")]

    all_violations = []

    for path in paths:
        if path.is_file() and path.suffix == ".py":
            violations = check_file(path)
            all_violations.extend(violations)
        elif path.is_dir():
            for py_file in path.rglob("*.py"):
                violations = check_file(py_file)
                all_violations.extend(violations)

    # Report violations
    if all_violations:
        for violation in all_violations:
            print(
                f"{violation.filename}:{violation.line}:{violation.column}: {violation.rule_code} {violation.message}"
            )
        sys.exit(1)
    else:
        print("No httpx.Client() violations found.")
        sys.exit(0)


if __name__ == "__main__":
    main()



from uipath.eval.evaluators import (
    BaseEvaluationCriteria,
    BaseEvaluator,
    BaseEvaluatorConfig,
)
from uipath.eval.models import AgentExecution, EvaluationResult, NumericEvaluationResult


class CSVShapeEvaluationCriteria(BaseEvaluationCriteria):
    """Evaluation criteria for the CSV shape evaluator."""

    expected_rows: int
    expected_columns: int


class CSVShapeEvaluatorConfig(BaseEvaluatorConfig[CSVShapeEvaluationCriteria]):
    """Configuration for the CSV shape evaluator."""

    name: str = "CSVShapeEvaluator"
    default_evaluation_criteria: CSVShapeEvaluationCriteria = (
        CSVShapeEvaluationCriteria(expected_rows=1, expected_columns=1)
    )


class CSVShapeEvaluator(
    BaseEvaluator[CSVShapeEvaluationCriteria, CSVShapeEvaluatorConfig, None]
):
    """A custom evaluator that checks if the CSV shape information is correct in the output attachment."""

    @classmethod
    def get_evaluator_id(cls) -> str:
        return "CSVShapeEvaluator"

    async def evaluate(
        self,
        agent_execution: AgentExecution,
        evaluation_criteria: CSVShapeEvaluationCriteria,
    ) -> EvaluationResult:
        # The agent prints: "CSV shape (rows, columns)\nCSV columns [...]"
        # We need to find this in the captured output

        expected_shape = f"({evaluation_criteria.expected_rows}, {evaluation_criteria.expected_columns})"
        shape_found = False

        # Check agent traces (where print output is captured)
        for span in agent_execution.agent_trace:
            # Check span attributes
            if span.attributes:
                for attr_value in span.attributes.values():
                    if isinstance(attr_value, str) and expected_shape in attr_value:
                        shape_found = True
                        break

            # Check span events (where stdout might be captured)
            if hasattr(span, 'events') and span.events:
                for event in span.events:
                    if hasattr(event, 'attributes') and event.attributes:
                        for attr_value in event.attributes.values():
                            if isinstance(attr_value, str) and expected_shape in attr_value:
                                shape_found = True
                                break

            if shape_found:
                break

        # Check agent output
        if not shape_found and agent_execution.agent_output:
            output_str = str(agent_execution.agent_output)
            shape_found = expected_shape in output_str

        return NumericEvaluationResult(
            score=float(shape_found),
        )



import importlib.util
import logging
import sys
from pathlib import Path
from typing import Any

from pydantic import TypeAdapter

from uipath._cli._evals._helpers import (  # type: ignore # Remove after gnarly fix
    try_extract_file_and_class_name,
)
from uipath._cli._evals._models._evaluator import CodedEvaluator, LegacyEvaluator
from uipath._utils.constants import EVALS_FOLDER
from uipath.eval.evaluators import (
    BaseEvaluator,
    BaseLegacyEvaluator,
    LegacyContextPrecisionEvaluator,
    LegacyFaithfulnessEvaluator,
    LegacyLlmAsAJudgeEvaluator,
    LegacyTrajectoryEvaluator,
)
from uipath.eval.evaluators.base_evaluator import GenericBaseEvaluator

logger = logging.getLogger(__name__)


class EvaluatorFactory:
    """Factory class for creating evaluator instances based on configuration."""

    @staticmethod
    def _prepare_evaluator_config(data: dict[str, Any]) -> dict[str, Any]:
        """Prepare evaluator config by merging top-level fields into evaluatorConfig.

        This allows flexibility in specifying fields like 'name' and 'description' either at the
        top level or within evaluatorConfig. Top-level values take precedence if both exist.

        Args:
            data: The raw evaluator data dictionary

        Returns:
            The prepared evaluatorConfig with merged fields
        """
        evaluator_config = data.get("evaluatorConfig", {})
        if not isinstance(evaluator_config, dict):
            evaluator_config = {}
        else:
            # Create a copy to avoid modifying the original
            evaluator_config = evaluator_config.copy()

        # Merge top-level 'name' into config if present
        if "name" in data and data["name"] is not None:
            # Top-level name takes precedence
            evaluator_config["name"] = data["name"]

        # Merge top-level 'description' into config if present
        if "description" in data and data["description"] is not None:
            # Top-level description takes precedence
            evaluator_config["description"] = data["description"]

        return evaluator_config

    @classmethod
    def create_evaluator(
        cls,
        data: dict[str, Any],
        evaluators_dir: Path | None = None,
        agent_model: str | None = None,
    ) -> GenericBaseEvaluator[Any, Any, Any]:
        if data.get("version", None) == "1.0":
            return cls._create_evaluator_internal(data, evaluators_dir)
        else:
            return cls._create_legacy_evaluator_internal(data, agent_model)

    @staticmethod
    def _create_evaluator_internal(
        data: dict[str, Any],
        evaluators_dir: Path | None = None,
    ) -> BaseEvaluator[Any, Any, Any]:
        # check custom evaluator
        evaluator_schema = data.get("evaluatorSchema", "")
        success, file_path, class_name = try_extract_file_and_class_name(
            evaluator_schema
        )
        if success:
            return EvaluatorFactory._create_custom_coded_evaluator_internal(
                data, file_path, class_name, evaluators_dir
            )
        else:
            return TypeAdapter(CodedEvaluator).validate_python(data)

    @staticmethod
    def _create_custom_coded_evaluator_internal(
        data: dict[str, Any],
        file_path_str: str,
        class_name: str,
        evaluators_dir: Path | None = None,
    ) -> BaseEvaluator[Any, Any, Any]:
        """Create a custom coded evaluator by dynamically loading from a Python file.

        Args:
            data: Dictionary containing evaluator configuration with evaluatorTypeId
                  in format "file://path/to/file.py:ClassName"
            evaluators_dir: Directory containing evaluator configuration files

        Returns:
            Instance of the dynamically loaded evaluator class

        Raises:
            ValueError: If file or class cannot be loaded, or if the class is not a BaseEvaluator subclass
        """
        file_path = Path(file_path_str)
        if not file_path.is_absolute():
            if not file_path.exists():
                if evaluators_dir is not None:
                    # Try the file directly in evaluators_dir first
                    file_path = evaluators_dir / file_path_str
                    if not file_path.exists():
                        # Fall back to evaluators_dir/custom/
                        file_path = evaluators_dir / "custom" / file_path_str
                else:
                    # Fall back to the old behavior
                    file_path = (
                        Path.cwd()
                        / EVALS_FOLDER
                        / "evaluators"
                        / "custom"
                        / file_path_str
                    )

        if not file_path.exists():
            raise ValueError(
                f"Evaluator file not found: {file_path}. "
                f"Make sure the file exists in the evaluators/custom/ directory"
            )

        module_name = f"_custom_evaluator_{file_path.stem}_{id(data)}"
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        if spec is None or spec.loader is None:
            raise ValueError(f"Could not load module from {file_path}")

        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        try:
            spec.loader.exec_module(module)
        except Exception as e:
            raise ValueError(
                f"Error executing module from {file_path}: {str(e)}"
            ) from e

        # Get the class from the module
        if not hasattr(module, class_name):
            raise ValueError(
                f"Class '{class_name}' not found in {file_path}. "
                f"Available classes: {[name for name in dir(module) if not name.startswith('_')]}"
            )

        evaluator_class = getattr(module, class_name)

        if not isinstance(evaluator_class, type) or not issubclass(
            evaluator_class, BaseEvaluator
        ):
            raise ValueError(
                f"Class '{class_name}' must be a subclass of BaseEvaluator"
            )

        evaluator_id = data.get("id")
        if not evaluator_id or not isinstance(evaluator_id, str):
            raise ValueError("Evaluator 'id' must be a non-empty string")
        return TypeAdapter(evaluator_class).validate_python(data)

    @staticmethod
    def _create_legacy_evaluator_internal(
        data: dict[str, Any],
        agent_model: str | None = None,
    ) -> BaseLegacyEvaluator[Any]:
        """Create an evaluator instance from configuration data.

        Args:
            data: Dictionary containing evaluator configuration from JSON file
            agent_model: Optional model name from agent settings for resolving
                'same-as-agent' model configuration

        Returns:
            Appropriate evaluator instance based on category

        Raises:
            ValueError: If category is unknown or required fields are missing
        """
        evaluator: LegacyEvaluator = TypeAdapter(LegacyEvaluator).validate_python(data)

        if isinstance(
            evaluator,
            LegacyTrajectoryEvaluator
            | LegacyLlmAsAJudgeEvaluator
            | LegacyContextPrecisionEvaluator
            | LegacyFaithfulnessEvaluator,
        ):
            if evaluator.model == "same-as-agent":
                if not agent_model:
                    raise ValueError(
                        "'same-as-agent' model option requires agent settings. "
                        "Ensure agent.json contains valid model settings."
                    )
                logger.info(
                    f"Resolving 'same-as-agent' to agent model: {agent_model} "
                    f"for evaluator '{evaluator.name}'"
                )
                evaluator.model = agent_model

        return evaluator
